# cython: language_level=3, embedsignature=True
# distutils: language=c++
"""
A `ParticleArray` represents a collection of particles.

"""

# logging imports
import logging
logger = logging.getLogger()

# numpy imports
cimport numpy
import numpy

# PyZoltan imports
from cyarray.carray cimport *

# Local imports
from cpython cimport PyObject
from cpython cimport *
from cython cimport *

from compyle.config import get_config
from compyle.array import Array, get_backend, to_device
from pysph.base.device_helper import DeviceHelper

# Maximum value of an unsigned int
cdef extern from "limits.h":
    cdef unsigned int UINT_MAX

_UINT_MAX = UINT_MAX

# Declares various tags for particles, and functions to check them.

# Note that these tags are the ones set in the 'tag' property of the
# particles, in a particle array. To define additional discrete
# properties, one can add another integer property to the particles in
# the particle array while creating them.

# These tags could be considered as 'system tags' used internally to
# distinguish among different kinds of particles. If more tags are
# needed for a particular application, add them as mentioned above.

# The is_* functions defined below are to be used in Python for tests
# etc. Cython modules can directly use the enum name.

cpdef bint is_local(int tag):
    return tag == Local

cpdef bint is_remote(int tag):
    return tag == Remote

cpdef bint is_ghost(int tag):
    return tag == Ghost

cpdef int get_local_tag():
    return Local

cpdef int get_remote_tag():
    return Remote

cpdef int get_ghost_tag():
    return Ghost

cdef class ParticleArray:
    """
    Class to represent a collection of particles.

    Attributes
    ----------
    name : str
        name of this particle array.
    properties : dict
        dictionary of {prop_name:carray}.
    constants : dict
        dictionary of {const_name: carray}

    Examples
    --------

    There are many ways to create a ParticleArray::

        >>> p = ParticleArray(name='fluid', x=[1.,2., 3., 4.])
        >>> p.name
        'fluid'
        >>> p.x, p.tag, p.pid, p.gid

    For a full specification of properties with their type etc.::

        >>> p = ParticleArray(name='fluid',
        ...                   x=dict(data=[1,2], type='int', default=1))
        >>> p.get_carray('x').get_c_type()
        'int'

    The default value is what is used to set the default value when a new
    particle is added and the arrays auto-resized.

    To create constants that are not resized with added/removed particles::

        >>> p = ParticleArray(name='f', x=[1,2], constants={'c':[0.,0.,0.]})

    """
    ######################################################################
    # `object` interface
    ######################################################################
    def __init__(self, str name='', default_particle_tag=Local,
                 constants=None, backend=None, **props):
        """Constructor

        Parameters
        ----------

        name : str
            name of this particle array.

        default_particle_tag : int
            one of `Local`, `Remote` or `Ghost`

        constants : dict
            dictionary of constant arrays for the entire particle array.  These
            must be arrays and are not resized when particles are added or
            removed.  These are stored as CArrays internally.

        props :
            any additional keyword arguments are taken to be properties, one
            for each property.

        """
        self.backend = get_backend(backend)
        self.time = 0.0
        self.name = name

        self.properties = {}
        self.default_values = {'tag':default_particle_tag}
        self.stride = {}

        self._initialize(**props)

        self.constants = {}
        if constants is not None:
            for const, data in constants.items():
                self.add_constant(name=const, data=data)

        # default lb_props are all the arrays
        self.lb_props = None

        # list of output property arrays
        self.output_property_arrays = []

        if self.backend is not 'cython':
            h = DeviceHelper(self, backend=self.backend)
            self.set_device_helper(h)
        else:
            self.gpu = None

    def __getattr__(self, name):
        """Convenience, to access particle property arrays as an attribute

        A numpy array is returned. Look at the get() functions documentation for
        care on using this numpy array.

        """
        if name in self.properties:
            return self._get_real_particle_prop(name)
        elif name in self.constants:
            return self.constants[name].get_npy_array()
        else:
            msg = "ParticleArray %s has no property/constant %s."\
                    %(self.name, name)
            raise AttributeError(msg)

    def __setattr__(self, name, value):
        """Convenience, to set particle property arrays as an attribute """
        self.set(**{name:value})

    def __reduce__(self):
        """Implemented to facilitate pickling of extension types """
        d = {}
        d['name'] = self.name
        props = {}
        default_values = {}

        for prop, arr in self.properties.items():
            pinfo = {}
            pinfo['name'] = prop
            pinfo['type'] = arr.get_c_type()
            pinfo['data'] = arr.get_npy_array()
            pinfo['default'] = self.default_values[prop]
            pinfo['stride'] = self.stride.get(prop, 1)
            props[prop] = pinfo

        d['properties'] = props

        props = {}
        for prop, arr in self.constants.items():
            pinfo = dict(name=prop, data=arr)
            props[prop] = pinfo

        d['constants'] = props

        return (ParticleArray, (), d)

    def __setstate__(self, d):
        """ Load the particle array object from the saved dictionary """
        self.properties = {}
        self.constants = {}
        self.property_arrays = []
        self.default_values = {}
        self.num_real_particles = 0

        self.name = d['name']
        props = d['properties']
        for prop in props:
            self.add_property(**props[prop])

        consts = d['constants']
        for prop in consts:
            self.add_constant(**consts[prop])

        self.num_real_particles = numpy.sum(props['tag']['data']==Local)

    def _initialize(self, **props):
        """Initialize the particle array with the given properties.

        Parameters
        ----------

        props : dict
            dictionary containing various property arrays. All these
            arrays are expected to be numpy arrays or objects that can
            be converted to numpy arrays.

        Notes
        -----

         - This will clear any existing data.
         - As a rule internal arrays will always be either long or double
           arrays.

        """
        cdef int nprop, nparticles
        cdef bint tag_present = False
        cdef numpy.ndarray a, arr, npyarr
        cdef IntArray tagarr
        cdef str name

        self.clear()

        nprop = len(props)

        if nprop == 0:
            return

        # Iterate over all props to find the maximum number of values passed.
        nv = 0
        _props = {}
        for name, prop in props.items():
            if isinstance(prop, dict):
                if 'data' in prop:
                    d = prop['data']
                    stride = prop.get('stride', 1)
                    if d is not None:
                        d = numpy.ravel(prop['data'])
                        prop['data'] = d
                        nv = max(nv, len(d)//stride)
                _props[name] = prop
            elif prop is not None:
                d = numpy.ravel(prop)
                _props[name] = d
                nv = max(nv, len(d))
            else:
                _props[name] = prop
        props.update(_props)

        # add the properties
        for name, prop in props.items():
            if isinstance(prop, dict):
                prop_info = prop
                prop_info['name'] = name
                data = prop.get('data', None)
                if data is not None:
                    if nv > 1 and len(prop['data']) == 1:
                        prop_info['data'] = numpy.ones(nv)*data
            else:
                if nv > 1 and len(prop) == 1:
                    prop = numpy.ones(nv)*prop
                prop_info = dict(name=name, data=prop)
            self.add_property(**prop_info)

        self.align_particles()

    ######################################################################
    # `Public` interface
    ######################################################################
    def update_backend(self, backend=None):
        self.backend = get_backend(backend)

    def set_output_arrays(self, list props):
        """Set the list of output arrays for this ParticleArray

        Parameters
        ----------

        props : list
            The list of property arrays

        Notes
        -----

        In PySPH, the solver obtains the list of property arrays to
        output by calling the `ParticleArray.get_property_arrays`
        method. If detailed output is not requested, the
        `output_property_arrays` attribute is used to determine the
        arrays that will be written to file

        """
        # first check if the arrays are valid and raise a warning
        for prop in props:
            self._check_property(prop)

        self.output_property_arrays = props

    def add_output_arrays(self, list props):
        """Append props to the existing list of output arrays

        Parameters
        ----------

        props : list
            The additional list of property arrays to save

        """
        # first check if the arrays are valid and raise a warning
        for prop in props:
            self._check_property(prop)

        # add to the existing list
        self.output_property_arrays.extend(props)
        self.output_property_arrays = list( set(self.output_property_arrays) )

    def get_property_arrays(self, all=True, only_real=True):
        """Return a dictionary of arrays held by the `ParticleArray` container.

        This does not include the constants.

        Parameters
        ----------

        all : bint
            Flag to select all arrays

        only_real : bint
            Flag to select Local/Remote particles

        Notes
        -----

        The dictionary returned is keyed on the property name and the
        value is the NumPy array representing the data. If `all` is
        set to False, the list of arrays is determined by the
        `output_property_arrays` data attribute.

        """
        # the arrays dictionary to be returned
        arrays = {}

        # the list of properties
        props = self.output_property_arrays
        if all or len(props) == 0:
            props = list(self.properties.keys())

        # number of particles
        num_particles = self.get_number_of_particles(only_real)
        if self.gpu is not None and self.backend is not 'cython':
            self.gpu.pull(*props)

        # add the property arrays
        for prop in props:
            stride = self.stride.get(prop, 1)
            prop_array = self.properties[prop].get_npy_array()
            arrays[prop] = prop_array[:num_particles*stride]

        return arrays

    cpdef set_num_real_particles(self, long value):
        self.num_real_particles = value

    cpdef has_array(self, str arr_name):
        """Returns true if the array arr_name is present """
        return self.properties.has_key(arr_name)

    def clear(self):
        """Clear all data held by this array """
        self.properties = {'tag':IntArray(0), 'pid':IntArray(0), 'gid':UIntArray(0)}
        tag_def_values = self.default_values['tag']
        self.default_values.clear()
        self.default_values = {'tag':tag_def_values, 'pid':0, 'gid':_UINT_MAX}

    cpdef set_time(self, double time):
        self.time = time

    cpdef double get_time(self):
        return self.time

    cpdef set_name(self, str name):
        self.name = name

    def set_lb_props(self, list lb_props):
        self.lb_props = lb_props

    cpdef get_lb_props(self):
        """Return the properties that are to be load balanced.  If none are
        explicitly set by the user, return all of the properties.
        """
        if self.lb_props is None:
            return list(self.properties.keys())
        else:
            return self.lb_props

    cpdef int get_number_of_particles(self, bint real=False):
        """ Return the number of particles """
        if self.gpu is not None and self.backend is not 'cython':
            return self.gpu.get_number_of_particles()
        if real:
            return self.num_real_particles
        else:
            if 'tag' in self.properties:
                return self.properties['tag'].length
            elif len(self.properties) > 0:
                prop = list(self.properties.keys())[0]
                stride = self.stride.get(prop, 1)
                return self.properties[prop].length//stride
            else:
                return 0

    cpdef remove_particles(self, indices, align=True):
        """ Remove particles whose indices are given in index_list.

        We repeatedly interchange the values of the last element and values from
        the index_list and reduce the size of the array by one. This is done for
        every property that is being maintained.

        Parameters
        ----------

        indices : array
            an array of indices, this array can be a list, numpy array
            or a LongArray.

        Notes
        -----

        Pseudo-code for the implementation::

            if index_list.length > number of particles
                raise ValueError

            sorted_indices <- index_list sorted in ascending order.

            for every every array in property_array
                array.remove(sorted_indices)

        """
        if self.gpu is not None and self.backend is not 'cython':
            if type(indices) != Array:
                if isinstance(indices, BaseArray):
                    indices = indices.get_npy_array()
                else:
                    indices = numpy.asarray(indices)
                indices = to_device(indices.astype(numpy.uint32),
                                    backend=self.backend)
            return self.gpu.remove_particles(indices, align=align)

        cdef BaseArray index_list
        if isinstance(indices, BaseArray):
            index_list = indices
        else:
            indices = numpy.asarray(indices)
            index_list = LongArray(indices.size)
            index_list.set_data(indices)

        cdef str msg
        cdef numpy.ndarray sorted_indices
        cdef BaseArray prop_array
        cdef int num_arrays, i
        cdef list property_arrays

        if index_list.length > self.get_number_of_particles():
            msg = 'Number of particles to be removed is greater than'
            msg += 'number of particles in array'
            raise ValueError(msg)

        sorted_indices = numpy.sort(index_list.get_npy_array())
        num_arrays = len(self.properties)

        for name, prop_array in self.properties.items():
            stride = self.stride.get(name, 1)
            prop_array.remove(sorted_indices, 1, stride)

        if index_list.length > 0 and align:
            self.align_particles()

    cpdef remove_tagged_particles(self, int tag, bint align=True):
        """ Remove particles that have the given tag.

        Parameters
        ----------

        tag : int
            the type of particles that need to be removed.

        """
        if self.gpu is not None and self.backend is not 'cython':
            return self.gpu.remove_tagged_particles(tag, align=align)
        cdef LongArray indices = LongArray()
        cdef IntArray tag_array = self.properties['tag']
        cdef int *tagarrptr = tag_array.get_data_ptr()
        cdef int i

        # find the indices of the particles to be removed.
        for i in range(tag_array.length):
            if tagarrptr[i] == tag:
                indices.append(i)

        # remove the particles.
        self.remove_particles(indices, align=align)

    def add_particles(self, align=True, **particle_props):
        """
        Add particles in particle_array to self.

        Parameters
        ----------

        particle_props : dict
            a dictionary containing numpy arrays for various particle
            properties.

        Notes
        -----

         - all properties should have same length arrays.
         - all properties should already be present in this particles array.
           if new properties are seen, an exception will be raised.

        """
        cdef str prop
        cdef BaseArray arr

        if len(particle_props) == 0:
            return 0

        # check if the input properties are valid.
        prop = ''
        for prop in particle_props:
            self._check_property(prop)

        if self.gpu is not None and self.backend is not 'cython':
            gpu_particle_props = {}
            for prop, ary in particle_props.items():
                if prop in self.gpu.properties:
                    dtype = self.gpu.get_device_array(prop).dtype
                else:
                    dtype = self.default_values[prop]
                gpu_particle_props[prop] = to_device(
                        numpy.array(ary, dtype=dtype),
                        backend=self.backend)
            return self.gpu.add_particles(align=align, **gpu_particle_props)

        cdef int num_extra_particles, old_num_particles, new_num_particles
        cdef numpy.ndarray s_arr, nparr

        if len(prop) == 0:
            num_extra_particles = 0
        else:
            stride = self.stride.get(prop, 1)
            num_extra_particles = len(particle_props[prop])//stride
        old_num_particles = self.get_number_of_particles()
        new_num_particles = num_extra_particles + old_num_particles

        for prop in self.properties:
            arr = <BaseArray>PyDict_GetItem(self.properties, prop)
            stride = self.stride.get(prop, 1)

            if PyDict_Contains(particle_props, prop)== 1:
                d_type = arr.get_npy_array().dtype
                s_arr = numpy.asarray(particle_props[prop], dtype=d_type)
                arr.extend(s_arr)
            else:
                arr.resize(new_num_particles*stride)
                # set the properties of the new particles to the default ones.
                nparr = arr.get_npy_array()
                nparr[old_num_particles*stride:] = self.default_values[prop]

        if num_extra_particles > 0 and align:
            # make sure particles are aligned properly.
            self.align_particles()

        return 0

    cpdef int append_parray(self, ParticleArray parray, bint align=True,
            bint update_constants=False) except -1:
        """ Add particles from a particle array

        properties that are not there in self will be added
        """
        if parray.get_number_of_particles() == 0:
            return 0

        if self.gpu is not None and self.backend is not 'cython':
            self.gpu.append_parray(parray, align=align,
                                   update_constants=update_constants)
            return 0

        cdef int num_extra_particles = parray.get_number_of_particles()
        cdef int old_num_particles = self.get_number_of_particles()
        cdef int new_num_particles = num_extra_particles + old_num_particles
        cdef str prop_name
        cdef BaseArray arr, source, dest
        cdef numpy.ndarray nparr_dest, nparr_source

        # extend current arrays by the required number of particles
        self.extend(num_extra_particles)

        for prop_name in parray.properties:
            if PyDict_Contains(self.properties, prop_name):
                arr = <BaseArray>PyDict_GetItem(self.properties, prop_name)
            else:
                arr = None

            stride = self.stride.get(prop_name, 1)
            if arr is not None:
                source = <BaseArray>PyDict_GetItem(parray.properties, prop_name)
                nparr_source = source.get_npy_array()
                nparr_dest = arr.get_npy_array()
                nparr_dest[old_num_particles*stride:] = nparr_source
            else:
                # meaning this property is not there in self.
                stride = parray.stride.get(prop_name, 1)
                self.add_property(name=prop_name,
                                  type=parray.properties[prop_name].get_c_type(),
                                  default=parray.default_values[prop_name],
                                  stride=stride
                                  )
                # now add the values to the end of the created array
                dest = <BaseArray>PyDict_GetItem(self.properties, prop_name)
                nparr_dest = dest.get_npy_array()
                source = <BaseArray>PyDict_GetItem(parray.properties, prop_name)
                nparr_source = source.get_npy_array()
                nparr_dest[old_num_particles*stride:] = nparr_source

        if update_constants:
            for const in parray.constants:
                self.constants.setdefault(const, parray.constants[const])

        if num_extra_particles > 0 and align:
            self.align_particles()

        return 0

    cpdef extend(self, int num_particles):
        """ Increase the total number of particles by the requested amount

        New particles are added at the end of the list, you will have to manually
        call align_particles later in order to update the number of particles.
        """
        if num_particles <= 0:
            return

        if self.gpu is not None and self.backend is not 'cython':
            self.gpu.extend(num_particles)
            return 0

        cdef int old_size = self.get_number_of_particles()
        cdef int new_size = old_size + num_particles
        cdef BaseArray arr
        cdef numpy.ndarray nparr
        cdef int stride

        for key in self.properties:
            stride = self.stride.get(key, 1)
            arr = self.properties[key]
            arr.resize(new_size*stride)
            nparr = arr.get_npy_array()
            nparr[old_size*stride:] = self.default_values[key]

    cdef numpy.ndarray _get_real_particle_prop(self, str prop_name):
        """ get the npy array of property corresponding to only real particles

        No checks are performed. Only call this after making sure that the
        property required already exists.
        """
        cdef BaseArray prop_array
        prop_array = self.properties.get(prop_name)
        stride = self.stride.get(prop_name, 1)
        if prop_array is not None:
            return prop_array.get_npy_array()[:self.num_real_particles*stride]
        else:
            return None

    def get(self, *args, only_real_particles=True):
        """ Return the numpy array/constant for the  property names in
        the arguments.

        Parameters
        ----------

        only_real_particles : bool
            indicates if properties of only real particles need to be
            returned or all particles to be returned. By default only
            real particles will be returned.
         args : additional args
            a list of property names.

        Notes
        -----

        The returned numpy array does **NOT** own its data. Other
        operations may be performed.

        Returns
        -------

        Numpy array.

        """
        cdef int nargs = len(args)
        cdef list result = []
        cdef str arg
        cdef int i, stride
        cdef BaseArray arg_array

        if nargs == 0:
            return

        if only_real_particles == True:
            for i in range(nargs):
                arg = args[i]
                self._check_property(arg)
                stride = self.stride.get(arg, 1)
                if arg in self.properties:
                    arg_array = self.properties[arg]
                    result.append(
                        arg_array.get_npy_array()[:self.num_real_particles*stride])

                else:
                    result.append(self.constants[arg].get_npy_array())
        else:
            for i in range(nargs):
                arg = args[i]
                self._check_property(arg)

                if arg in self.properties:
                    arg_array = self.properties[arg]
                    result.append(arg_array.get_npy_array())
                else:
                    result.append(self.constants[arg].get_npy_array())

        if nargs == 1:
            return result[0]
        else:
            return tuple(result)

    def set_device_helper(self, gpu):
        """Set the device helper to push/pull from a hardware accelerator.
        """
        self.gpu = gpu

    def set(self, **props):
        """ Set properties from numpy arrays like objects

        Parameters
        ----------

        props : dict
            a dictionary of properties containing the arrays to be set.

        Notes
        -----

         - the properties being set must already be present in the properties
           dict.
         - the size of the data should match the array already present.

        """
        cdef str prop
        cdef BaseArray prop_array
        cdef int nprops = len(props)
        cdef list prop_names = list(props.keys())
        cdef int i

        for i in range(nprops):
            prop = prop_names[i]
            self._check_property(prop)

        for prop in props:
            proparr = numpy.asarray(props[prop])
            if self.properties.has_key(prop):
                prop_array = self.properties[prop]
                prop_array.set_data(proparr)
            elif self.constants.has_key(prop):
                prop_array = self.constants[prop]
                prop_array.set_data(proparr)

            # if the tag property is being set, the alignment will have to be
            # changed.

    cpdef BaseArray get_carray(self, str prop):
        """Return the c-array for the property or constant.
        """
        if PyDict_Contains(self.properties, prop) == 1:
            return <BaseArray>PyDict_GetItem(self.properties, prop)
        elif PyDict_Contains(self.constants, prop) == 1:
            return <BaseArray>PyDict_GetItem(self.constants, prop)
        else:
            raise KeyError(
                'Property/constant "%s" not present in particle array.' % prop
            )

    cpdef add_constant(self, str name, data):
        """Add a constant property to the particle array.

        A constant property is an array but has a fixed size in that it is
        never resized as particles are added or removed.  These properties
        are always stored internally as CArrays.

        An example of where this is useful is if you need to calculate the
        center of mass of a solid body or the net force on the body.

        Parameters
        ----------

        name : str
            name of the constant
        data : array-like
            the value for the data.
        """
        if name in self.constants:
            raise RuntimeError('Constant called "%s" already exists.'%name)
        if name in self.properties:
            raise RuntimeError('Property called "%s" already exists.'%name)

        array_data = numpy.ravel(data)
        self.constants[name] = self._create_c_array_from_npy_array(array_data)
        if self.gpu is not None:
            self.gpu.add_const(name, self.constants[name])

    cpdef add_property(self, str name, str type='double', default=None, data=None,
                       stride=1):
        """Add a new property to the particle array.

        If a `default` is not supplied 0 is assumed.  The stride is useful when
        many elements are needed per particle.  For example if stride is 3 then
        3 elements are allocated per particle.


        Parameters
        ----------

        name : str
            compulsory name of property.
        type : str
            specifying the data type of this property ('double', 'int'
            etc.)
        default : value
            specifying the default value of this property.
        data : ndarray
            specifying the data associated with each particle.
        stride : int
            the number of elements per particle.

        Notes
        -----

        If there are no particles currently in the particle array, and a
        new property with some particles is added, all the remaining
        properties will be resized to the size of the newly added array.

        If there are some particles in the particle array, and a new
        property is added without any particles, then this new property will
        be resized according to the current size.

        If there are some particles in the particle array and a new property
        is added with a different number of particles, then an error will be
        raised.

        Warning
        -------

            - it is best not to add properties with data when you already have
              particles in the particle array. Reason is that, the particles in
              the particle array will be stored so that the 'Real' particles are
              in the top of the list and later the dummy ones. The data in your
              property array should be matched to the particles appropriately.
              This may not always be possible when there are particles of
              different type in the particle array.
            - Properties without any values can be added anytime.
            - While initializing particle arrays only using the add_property
              function, you will have to call align_particles manually to make
              sure particles are aligned properly.

        """
        cdef str prop_name=name, data_type=type
        cdef bint array_size_proper = False

        if data is not None:
            try:
                len(data)
            except TypeError:
                data = numpy.ones(self.get_number_of_particles()*stride)*data
            else:
                data = numpy.ravel(data)

        # make sure the size of the supplied array is consistent.
        if (data is None or self.get_number_of_particles() == 0 or
            len(data) == 0):
            array_size_proper = True
        else:
            if (self.get_number_of_particles() == len(data)//stride) and \
               (len(data) % stride == 0):
                array_size_proper = True

        if array_size_proper == False:
            msg = 'Array sizes incompatible for property: %s' % name
            logger.error(msg)
            raise ValueError(msg)

        # setup the default values
        if default is None:
            if prop_name not in self.properties:
                default = 0
            else:
                default = self.default_values[prop_name]

        self.default_values[prop_name] = default
        if stride != 1:
            self.stride[name] = stride

        # array sizes are compatible, now resize the required arrays
        # appropriately and add.
        if self.get_number_of_particles() == 0:
            if data is None or len(data) == 0:
                # if a property with that name already exists, do not do
                # anything.
                if prop_name not in self.properties:
                    # just add the property with a zero array.
                    self.properties[prop_name] = self._create_carray(
                        data_type, 0, default)
            else:
                # new property has been added with some particles, while no
                # particles are currently present. First resize the current
                # properties to this new length, and then add this new
                # property.
                n_elem = len(data)//stride
                for prop in self.properties:
                    arr = self.properties[prop]
                    prop_stride = self.stride.get(prop, 1)
                    arr.resize(n_elem*prop_stride)
                    arr.get_npy_array()[:] = self.default_values[prop]
                if prop_name == 'tag':
                    arr = numpy.asarray(data)
                    self.num_real_particles = numpy.sum(arr==Local)
                else:
                    self.num_real_particles = n_elem

                if self.properties.has_key(prop_name):
                    # just add the particles to the already existing array.
                    d_type = self.properties[prop_name].get_npy_array().dtype
                    arr = numpy.asarray(data, dtype=d_type)
                    self.properties[prop_name].set_data(arr)
                else:
                    # now add the new property array
                    # if a type was specifed create that type of array
                    if data_type is None:
                        # get an array for this data
                        self.properties[prop_name] = (
                            self._create_c_array_from_npy_array(data))
                    else:
                        arr = self._create_carray(data_type, len(data), default)
                        np_arr = arr.get_npy_array()
                        np_arr[:] = data
                        self.properties[prop_name] = arr
        else:
            if data is None or len(data) == 0:
                # new property is added without any initial data, resize it to
                # current particle count.
                if not self.properties.has_key(prop_name):
                    arr = self._create_carray(
                        data_type, self.get_number_of_particles()*stride,
                        default
                    )
                    self.properties[prop_name] = arr
            else:
                if self.properties.has_key(prop_name):
                    d_type = self.properties[prop_name].get_npy_array().dtype
                    arr = numpy.asarray(data, dtype=d_type)
                    self.properties[prop_name].set_data(arr)
                else:
                    if data_type is None:
                        # just add the property array
                        self.properties[prop_name] = (
                            self._create_c_array_from_npy_array(data))
                    else:
                        arr = self._create_carray(data_type, len(data), default)
                        np_arr = arr.get_npy_array()
                        arr.get_npy_array()[:] = data
                        self.properties[prop_name] = arr

        if self.gpu is not None:
            self.gpu.add_prop(prop_name, self.properties[prop_name])
            if self.gpu.get_number_of_particles() == 0:
                self.gpu.push()

    ######################################################################
    # Non-public interface
    ######################################################################
    def _create_carray(self, str data_type, int size, default=0):
        """Create a carray of the requested type, and of requested size

        Parameters
        ----------

        data_type : str
            string representing the 'c' data type - eg. 'int' for
            integers.
        size : int
            the size of the requested array
        default : value
            the default value to initialize the array with.

        """
        cdef BaseArray arr
        if data_type == None:
            arr = DoubleArray(size)
        elif data_type == 'double':
            arr = DoubleArray(size)
        elif data_type == 'long':
            arr = LongArray(size)
        elif data_type == 'float':
            arr = FloatArray(size)
        elif data_type == 'int':
            arr = IntArray(size)
        elif data_type == 'unsigned int':
            arr = UIntArray(size)
        else:
            logger.error('Trying to create carray of unknown '
                   'datatype: %s' %data_type)

        if size > 0:
            arr.get_npy_array()[:] = default

        return arr

    cdef _check_property(self, str prop):
        """Check if a property is present or not """
        if (PyDict_Contains(self.properties, prop) or
            PyDict_Contains(self.constants, prop)):
            return
        else:
            raise AttributeError, 'property %s not present'%(prop)

    cdef object _create_c_array_from_npy_array(self, numpy.ndarray np_array):
        """Create and return  a carray array from the given numpy array.

        Notes
        -----

         - this function is used only when a C array needs to be
           created (in the initialize function).

        """
        cdef int np = np_array.size
        cdef object a
        if np_array.dtype == numpy.int32 or np_array.dtype == numpy.int64:
            a = LongArray(np)
            a.set_data(np_array)
        elif np_array.dtype == numpy.float32:
            a = FloatArray(np)
            a.set_data(np_array)
        elif np_array.dtype == numpy.double:
            a = DoubleArray(np)
            a.set_data(np_array)
        else:
            msg = 'unknown numpy data type passed %s'%(np_array.dtype)
            raise TypeError, msg

        return a

    cpdef int align_particles(self) except -1:
        """Moves all 'Local' particles to the beginning of the array

        This makes retrieving numpy slices of properties of 'Local'
        particles possible. This facility will be required frequently.

        Notes
        -----

        Pseudo-code::

            index_arr = LongArray(n)

            next_insert = 0
            for i from 0 to n
                p <- ith particle
                if p is Local
                    if i != next_insert
                        tmp = index_arr[next_insert]
                        index_arr[next_insert] = i
                        index_arr[i] = tmp
                        next_insert += 1
                    else
                        index_arr[i] = i
                        next_insert += 1
                else
                    index_arr[i] = i

             # we now have the new index assignment.
             # swap the required values as needed.
             for every property array:
                 for i from 0 to n:
                     if index_arr[i] != i:
                         tmp = prop[i]
                         prop[i] = prop[index_arr[i]]
                         prop[index_arr[i]] = tmp
        """
        if self.gpu is not None and self.backend is not 'cython':
            self.gpu.align_particles()
            return 0

        cdef size_t i, num_particles
        cdef size_t next_insert
        cdef int tmp
        cdef IntArray tag_arr
        cdef LongArray index_array
        cdef BaseArray arr
        cdef long num_real_particles = 0
        cdef long num_moves = 0

        next_insert = 0
        num_particles = self.get_number_of_particles()

        tag_arr = self.get_carray('tag')

        # malloc the new index array
        index_array = LongArray(num_particles)

        for i in range(num_particles):
            if tag_arr.data[i] == Local:
                num_real_particles += 1
                if i != next_insert:
                    tmp = index_array.data[next_insert]
                    index_array.data[next_insert] = i
                    index_array.data[i] = tmp
                    next_insert += 1
                    num_moves += 1
                else:
                    index_array.data[i] = i
                    next_insert += 1
            else:
                index_array.data[i] = i

        self.num_real_particles = num_real_particles
        # we now have the aligned indices. Rearrange the particles
        # accordingly.

        if num_moves > 0:
            for name, arr in self.properties.items():
                stride = self.stride.get(name, 1)
                arr.c_align_array(index_array, stride)

    cpdef ParticleArray empty_clone(self, props=None):
        """Creates an empty clone of the particle array
        """
        if self.gpu is not None and self.backend is not 'cython':
            return self.gpu.empty_clone(props=props)

        cdef ParticleArray result_array = ParticleArray()
        cdef list output_arrays
        cdef BaseArray src_arr
        cdef int stride
        cdef str prop_type, prop_name

        if props is None:
            prop_names = self.properties
        else:
            prop_names = props

        for const in self.constants:
            result_array.add_constant(const, data=self.constants[const])

        for prop_name in prop_names:
            prop_type = self.properties[prop_name].get_c_type()
            prop_default = self.default_values[prop_name]
            stride = self.stride.get(prop_name, 1)
            result_array.add_property(name=prop_name,
                                      type=prop_type,
                                      default=prop_default,
                                      stride=stride)

        result_array.set_name(self.name)

        if props is None:
            output_arrays = list(self.output_property_arrays)
        else:
            output_arrays = list(
                set(props).intersection(
                    self.output_property_arrays
                )
            )

        result_array.set_output_arrays(output_arrays)
        return result_array

    cpdef ensure_properties(self, ParticleArray src, list props=None):
        """Ensure that the particle array has the same properties as the
        one given.

        Note that this does not check for any constants but only properties.

        If the optional props argument is passed it only checks for these.
        """
        prop_names = props if props else src.properties.keys()

        for prop_name in prop_names:
            if prop_name not in self.properties:
                prop_type = src.properties[prop_name].get_c_type()
                prop_default = src.default_values[prop_name]
                stride = src.stride.get(prop_name, 1)
                self.add_property(
                    name=prop_name, type=prop_type,
                    default=prop_default, stride=stride
                )

    cpdef ParticleArray extract_particles(self, indices,
            ParticleArray dest_array=None, bint align=True, list props=None):
        """Create new particle array for particles with indices in index_array

        Parameters
        ----------

        indices : list/array/LongArray
            indices of particles to be extracted (can be a LongArray or
            list/numpy array).

        dest_array: ParticleArray
            optional Particle array to populate. Note that this array should
            have the necessary properties. If none is passed a new particle
            array is created and returned.

        align: bool
            Specify if the destination particle array is to be aligned after
            particle extraction.

        props : list
            the list of properties to extract, if None all properties
            are extracted.

        Notes
        -----

        The algorithm is as follows:

             - create a new particle array with the required properties.
             - resize the new array to the desired length (index_array.length)
             - copy the properties from the existing array to the new array.

        """
        if self.gpu is not None and self.backend is not 'cython':
            if type(indices) != Array:
                indices = to_device(
                        numpy.array(indices, dtype=numpy.uint32),
                        backend=self.backend)
            return self.gpu.extract_particles(indices, dest_array=dest_array,
                                              align=align, props=props)

        if not dest_array:
            dest_array = self.empty_clone(props=props)

        cdef BaseArray index_array
        if isinstance(indices, BaseArray):
            index_array = indices
        else:
            indices = numpy.asarray(indices)
            index_array = LongArray(indices.size)
            index_array.set_data(indices)

        cdef list prop_names, output_arrays
        cdef BaseArray dst_prop_array, src_prop_array
        cdef str prop_type, prop
        cdef int stride, start_idx

        if props is None:
            prop_names = list(self.properties.keys())
        else:
            prop_names = props

        # now we have the result array setup.
        # resize it
        if index_array.length == 0:
            return dest_array

        start_idx = dest_array.get_number_of_particles()

        dest_array.extend(index_array.length)

        # copy the required indices for each property.
        for prop in prop_names:
            src_prop_array = self.get_carray(prop)
            dst_prop_array = dest_array.get_carray(prop)
            stride = self.stride.get(prop, 1)
            src_prop_array.copy_values(index_array, dst_prop_array,
                                       stride, stride*start_idx)

        if align:
            dest_array.align_particles()

        return dest_array

    cpdef set_tag(self, long tag_value, LongArray indices):
        """Set value of tag to tag_value for the particles in indices """
        cdef LongArray tag_array = self.get_carray('tag')
        cdef int i

        for i in range(indices.length):
            tag_array.data[indices.data[i]] = tag_value

    cpdef copy_properties(self, ParticleArray source, long start_index=-1,
                          long end_index=-1):
        """ Copy properties from source to self

        Parameters
        ----------

        source : ParticleArray
            the particle array from where to copy.
        start_index : long
            the first particle in self which maps to the 0th particle in
            source
        end_index : long
            the index of first particle from start_index that
            is not copied

        """
        cdef BaseArray src_array, dst_array
        for prop_name in source.properties:
            if prop_name in self.properties:
                src_array = source.get_carray(prop_name)
                dst_array = self.get_carray(prop_name)
                stride = self.stride.get(prop_name, 1)
                dst_array.copy_subset(src_array, start_index, end_index, stride)

    cpdef copy_over_properties(self, dict props):
        """ Copy the properties from one set to another.

        Parameters
        ----------

        props : dict
            A mapping between the properties to be copied.

        Examples
        --------

        To save the properties 'x' and 'y' to say 'x0' and 'y0'::

            >>> pa.copy_over_properties(props = {'x':'x0', 'y':'y0'}

        """
        cdef DoubleArray dst, src
        cdef str prop
        cdef int stride

        cdef long np = self.get_number_of_particles()
        cdef long i

        for prop in props:

            src = self.get_carray(prop)
            dst = self.get_carray(props[prop])
            stride = self.stride.get(prop, 1)

            for i in range(np*stride):
                dst.data[i] = src.data[i]

    cpdef set_to_zero(self, list props):

        cdef long np = self.get_number_of_particles()
        cdef long i
        cdef int stride

        cdef DoubleArray prop_arr
        cdef str prop

        for prop in props:
            prop_arr = self.get_carray(prop)
            stride = self.stride.get(prop, 1)

            for i in range(np*stride):
                prop_arr.data[i] = 0.0

    cpdef set_pid(self, int pid):
        """Set the processor id for all particles """
        cdef IntArray pid_arr = self.properties['pid']
        cdef long a

        for a in range(pid_arr.length):
            pid_arr.data[a] = pid

    cpdef remove_property(self, str prop_name):
        """Removes property prop_name from the particle array """

        if self.properties.has_key(prop_name):
            self.properties.pop(prop_name)
            self.default_values.pop(prop_name)
        if prop_name in self.output_property_arrays:
            self.output_property_arrays.remove(prop_name)
        if self.gpu is not None:
            return self.gpu.remove_prop(prop_name)

    def update_min_max(self, props=None):
        """Update the min,max values of all properties """
        if self.gpu is not None and self.backend is not 'cython':
            backend = self.gpu
        else:
            backend = self

        if props:
            for prop in props:
                array = backend.properties[prop]
                array.update_min_max()
        else:
            for array in backend.properties.values():
                array.update_min_max()

    cpdef resize(self, long size):
        """Resize all arrays to the new size.  Note that this does not
        update the number of particles, as this just resizes the internal arrays.
        To do that, you need to call `align_particles`.

        """
        if self.gpu is not None and self.backend is not 'cython':
            return self.gpu.resize(size)

        for prop, array in self.properties.items():
            stride = self.stride.get(prop, 1)
            array.resize(size*stride)


# End of ParticleArray class
##############################################################################



import numpy as np
from pysph.base.utils import get_particle_array_wcsph
from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application
from pysph.sph.equation import Group
from pysph.sph.basic_equations import XSPHCorrection, ContinuityEquation
from pysph.sph.wc.basic import TaitEOS, MomentumEquation
from pysph.solver.solver import Solver
from pysph.sph.integrator import PECIntegrator
from pysph.sph.integrator_step import WCSPHStep


class EllipticalDrop(Application):
    def create_particles(self):
        dx = 0.025
        x, y = np.mgrid[-1.05:1.05:dx, -1.05:1.05:dx]
        mask = x*x + y*y < 1
        x = x[mask]
        y = y[mask]
        rho = 1.0
        h = 1.3*dx
        m = rho*dx*dx
        pa = get_particle_array_wcsph(
            name='fluid', x=x, y=y, u=-100*x, v=100*y, rho=rho,
            m=m, h=h
        )
        return [pa]

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    TaitEOS(dest='fluid', sources=None, rho0=1.0,
                            c0=1400, gamma=7.0),
                ],
                real=False
            ),

            Group(
                equations=[
                    ContinuityEquation(dest='fluid',  sources=['fluid']),

                    MomentumEquation(dest='fluid', sources=['fluid'],
                                     alpha=0.1, beta=0.0, c0=1400),

                    XSPHCorrection(dest='fluid', sources=['fluid']),
                ]
            ),
        ]
        return equations

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = PECIntegrator(fluid=WCSPHStep())

        dt = 5e-6
        tf = 0.0076
        solver = Solver(
            kernel=kernel, dim=2, integrator=integrator,
            dt=dt, tf=tf
        )
        return solver


if __name__ == '__main__':
    app = EllipticalDrop(fname='ed_no_scheme')
    app.run()



from collections import defaultdict
import functools
import inspect
from textwrap import dedent
import types

from mako.template import Template
import numpy as np

from compyle.config import get_config
from .equation import get_array_names
from .integrator_cython_helper import IntegratorCythonHelper
from .acceleration_eval_gpu_helper import (
    get_kernel_definition, get_converter, profile_kernel, wrap_code,
    get_helper_code
)


class GPUIntegrator(object):
    """Does the actual work of calling the kernels for integration.
    """
    def __init__(self, helper, c_acceleration_eval):
        self.helper = helper
        self.acceleration_eval = c_acceleration_eval
        self.nnps = None
        self.parallel_manager = None
        self.integrator = helper.object
        self._post_stage_callback = None
        self._use_double = get_config().use_double
        self._setup_methods()

    def _setup_methods(self):
        """This sets up a few methods of this class.

        This is unfortunately a bit hacky right now and should be cleaned
        later.

        It creates the methods for the following:

        self.one_timestep: this is the same as the integrator's method.

        self.initialize, self.stage1 ... self.stagen are created based on the
        number of steppers.

        """
        code = self.helper.get_timestep_code()
        ns = {}
        exec(code, ns)
        self.one_timestep = types.MethodType(ns['one_timestep'], self)

        for method in self.helper.get_stepper_method_wrapper_names():
            setattr(self, method, functools.partial(self._do_stage, method))

    def _do_stage(self, method):
        # Call the appropriate kernels for either initialize/stage computation.
        call_info = self.helper.calls[method]
        py_call_info = self.helper.py_calls['py_' + method]
        dtype = np.float64 if self._use_double else np.float32
        extra_args = [np.asarray(self.t, dtype=dtype),
                      np.asarray(self.dt, dtype=dtype),
                      np.asarray(0, dtype=np.uint32)]
        # Call the py_{method} for each destination.
        for name, (py_meth, dest) in py_call_info.items():
            py_meth(dest, *(extra_args[:-1]))

        # Call the stage* method for each destination.
        for name, (call, args, dest) in call_info.items():
            n = dest.get_number_of_particles(real=True)
            args[1] = (n,)
            # For NP_MAX
            extra_args[-1][...] = n - 1
            # Compute the remaining arguments.
            rest = [x() for x in args[3:]]
            call(*(args[:3] + rest + extra_args))

    def set_nnps(self, nnps):
        self.nnps = nnps

    def set_parallel_manager(self, pm):
        self.parallel_manager = pm

    def set_post_stage_callback(self, callback):
        self._post_stage_callback = callback

    def compute_accelerations(self, index=0, update_nnps=True):
        self.integrator.compute_accelerations(index, update_nnps)

    def update_domain(self):
        self.integrator.update_domain()

    def do_post_stage(self, stage_dt, stage):
        """This is called after every stage of the integrator.

        Internally, this calls any post_stage_callback function that has
        been given to take suitable action.

        Parameters
        ----------

         - stage_dt : double: the timestep taken at this stage.

         - stage : int: the stage completed (starting from 1).
        """
        self.t = self.orig_t + stage_dt
        if self._post_stage_callback is not None:
            self._post_stage_callback(self.t, self.dt, stage)

    def step(self, t, dt):
        """Main step routine.
        """
        self.orig_t = t
        self.t = t
        self.dt = dt
        self.one_timestep(t, dt)


class CUDAIntegrator(GPUIntegrator):
    """Does the actual work of calling the kernels for integration.
    """
    def _do_stage(self, method):
        from pycuda.gpuarray import splay
        import pycuda.driver as drv
        # Call the appropriate kernels for either initialize/stage computation.
        call_info = self.helper.calls[method]
        py_call_info = self.helper.py_calls['py_' + method]
        dtype = np.float64 if self._use_double else np.float32
        extra_args = [np.asarray(self.t, dtype=dtype),
                      np.asarray(self.dt, dtype=dtype)]
        # Call the py_{method} for each destination.
        for name, (py_meth, dest) in py_call_info.items():
            py_meth(dest, *extra_args)

        # Call the stage* method for each destination.
        for name, (call, args, dest) in call_info.items():
            n = dest.get_number_of_particles(real=True)

            gs, ls = splay(n)
            gs, ls = int(gs[0]), int(ls[0])

            num_blocks = (n + ls - 1) // ls
            num_tpb = ls

            # Compute the remaining arguments.
            args = [x() for x in args[3:]]
            call(*(args + extra_args),
                 block=(num_tpb, 1, 1), grid=(num_blocks, 1))


class IntegratorGPUHelper(IntegratorCythonHelper):
    def __init__(self, integrator, acceleration_eval_helper):
        super(IntegratorGPUHelper, self).__init__(
            integrator, acceleration_eval_helper
        )
        self.backend = acceleration_eval_helper.backend
        self.py_data = defaultdict(dict)
        self.data = defaultdict(dict)
        self.py_calls = defaultdict(dict)
        self.calls = defaultdict(dict)
        self.program = None

    def _setup_call_data(self):
        array_map = self.acceleration_eval_helper._array_map
        q = self.acceleration_eval_helper._queue
        calls = self.calls
        py_calls = self.py_calls
        steppers = self.object.steppers
        for method, info in self.py_data.items():
            for dest_name in info:
                py_meth = getattr(steppers[dest_name], method)
                dest = array_map[dest_name]
                py_calls[method][dest] = (py_meth, dest)

        for method, info in self.data.items():
            for dest_name, (kernel, args) in info.items():
                dest = array_map[dest_name]

                # Note: This is done to do some late binding. Instead of
                # just directly storing the dest.gpu.x, we compute it on
                # the fly as the number of particles and the actual buffer
                # may change.
                if self.backend == 'opencl':
                    def _getter(dest_gpu, x):
                        return getattr(dest_gpu, x).dev.data
                elif self.backend == 'cuda':
                    def _getter(dest_gpu, x):
                        return getattr(dest_gpu, x).dev

                _args = [
                    functools.partial(_getter, dest.gpu, x[2:])
                    for x in args
                ]
                all_args = [q, None, None] + _args
                call = getattr(self.program, kernel)
                call = profile_kernel(call, self.backend)
                calls[method][dest] = (call, all_args, dest)

    def get_code(self):
        if self.object is not None:
            tpl = dedent("""
            // ------------------------------------------------------------
            // Integrator steppers.
            ${helper.get_stepper_code()}

            // ------------------------------------------------------------
            % for dest in sorted(helper.object.steppers.keys()):
            // Steppers for ${dest}
            % for method in helper.get_stepper_method_wrapper_names():
            <% helper.get_py_stage_code(dest, method) %>
            % if helper.has_stepper_loop(dest, method):
            ${helper.get_stepper_kernel(dest, method)}
            % endif
            % endfor
            % endfor
            // ------------------------------------------------------------
            """)
            template = Template(text=tpl)
            return template.render(helper=self)
        else:
            return ''

    def setup_compiled_module(self, module, acceleration_eval):
        # Create the compiled module.
        self.program = module
        self._setup_call_data()
        if self.backend == 'opencl':
            gpu_integrator = GPUIntegrator(self, acceleration_eval)
        elif self.backend == 'cuda':
            gpu_integrator = CUDAIntegrator(self, acceleration_eval)
        # Setup the integrator to use this compiled module.
        self.object.set_compiled_object(gpu_integrator)

    def get_py_stage_code(self, dest, method):
        stepper = self.object.steppers[dest]
        method = 'py_' + method
        if hasattr(stepper, method):
            self.py_data[method][dest] = dest

    def get_timestep_code(self):
        method = self.object.one_timestep
        return dedent(''.join(inspect.getsourcelines(method)[0]))

    def get_stepper_code(self):
        classes = {}
        helpers = []
        for stepper in self.object.steppers.values():
            cls = stepper.__class__.__name__
            classes[cls] = stepper
            if hasattr(stepper, '_get_helpers_'):
                for helper in stepper._get_helpers_():
                    if helper not in helpers:
                        helpers.append(helper)

        known_types = dict(self.acceleration_eval_helper.known_types)

        Converter = get_converter(self.acceleration_eval_helper.backend)
        code_gen = Converter(known_types=known_types)

        wrappers = get_helper_code(helpers, code_gen, self.backend)
        for cls in sorted(classes.keys()):
            wrappers.append(code_gen.parse_instance(classes[cls]))
        return '\n'.join(wrappers)

    def get_stepper_kernel(self, dest, method):
        kernel = '{method}_{dest}'.format(dest=dest, method=method)
        stepper = self.object.steppers.get(dest)
        cls = stepper.__class__.__name__
        args = self.get_args(dest, method)
        if 'self' in args:
            args.remove('self')
        s, d = get_array_names(args)

        all_args = self.acceleration_eval_helper._get_typed_args(
            sorted(d) + ['t', 'dt']
        )
        all_args.append('unsigned int NP_MAX')

        # All the steppers are essentially empty structs so we just pass 0 as
        # the stepper struct as it is not used at all. This simplifies things
        # as we do not need to generate structs and pass them around.
        code = [
            'int d_idx = GID_0 * LDIM_0 + LID_0;',
            '/* Guard for padded threads. */',
            'if (d_idx > NP_MAX) {return;};'
        ] + wrap_code(
            '{cls}_{method}({args});'.format(
                cls=cls, method=method,
                args=', '.join(['0'] + args)
            ), indent=''
        )

        body = '\n'.join(' '*4 + x for x in code)

        self.data[method][dest] = (kernel, sorted(d))

        sig = get_kernel_definition(kernel, all_args)
        return (
            '{sig}\n{{\n{body}\n}}\n'.format(
                sig=sig, body=body
            )
        )




class Tool(object):
    """A tool is typically an object that can be used to perform a
    specific task on the solver's pre_step/post_step or post_stage callbacks.
    This can be used for a variety of things.  For example, one could save a
    plot, print debug statistics or perform remeshing etc.

    To create a new tool, simply subclass this class and overload any of its
    desired methods.
    """

    def pre_step(self, solver):
        """If overloaded, this is called automatically before each integrator
        step.  The method is passed the solver instance.
        """
        pass

    def post_stage(self, current_time, dt, stage):
        """If overloaded, this is called automatically after each integrator
        stage, i.e. if the integrator is a two stage integrator it will be
        called after the first and second stages.

        The method is passed (current_time, dt, stage).  See the the
        `Integrator.one_timestep` methods for examples of how this is called.
        """
        pass

    def post_step(self, solver):
        """If overloaded, this is called automatically after each integrator
        step.  The method is passed the solver instance.
        """
        pass


class SimpleRemesher(Tool):
    """A simple tool to periodically remesh a given array of particles onto an
    initial set of points.
    """

    def __init__(self, app, array_name, props, freq=100, xi=None, yi=None,
                 zi=None, kernel=None, equations=None):
        """Constructor.

        Parameters
        ----------

        app : pysph.solver.application.Application
            The application instance.
        array_name: str
            Name of the particle array that needs to be remeshed.
        props : list(str)
            List of properties to interpolate.
        freq : int
            Frequency of remeshing operation.
        xi, yi, zi : ndarray
            Positions to remesh the properties onto.  If not specified they
            are taken from the particle arrays at the time of construction.
        kernel: any kernel from pysph.base.kernels

        equations: list or None
            Equations to use for the interpolation, passed to the interpolator.

        """
        from pysph.solver.utils import get_array_by_name
        self.app = app
        self.particles = app.particles
        self.array = get_array_by_name(self.particles, array_name)
        self.props = props
        if xi is None:
            xi = self.array.x
        if yi is None:
            yi = self.array.y
        if zi is None:
            zi = self.array.z
        self.xi, self.yi, self.zi = xi.copy(), yi.copy(), zi.copy()
        self.freq = freq
        from pysph.tools.interpolator import Interpolator
        if kernel is None:
            kernel = app.solver.kernel
        self.interp = Interpolator(
            self.particles, x=self.xi, y=self.yi, z=self.zi,
            kernel=kernel,
            domain_manager=app.create_domain(),
            equations=equations
        )

    def post_step(self, solver):
        if solver.count % self.freq == 0 and solver.count > 0:
            self.interp.nnps.update()
            data = dict(x=self.xi, y=self.yi, z=self.zi)
            for prop in self.props:
                data[prop] = self.interp.interpolate(prop)
            self.array.set(**data)
            self.interp.nnps.update_domain()


class DensityCorrection(Tool):
    """
    A tool to reinitialize the density of the fluid particles
    """

    def __init__(self, app, arr_names, corr='shepard', freq=10, kernel=None):
        """
        Parameters
        ----------

        app : pysph.solver.application.Application.
            The application instance.
        arr_names : array
            Names of the particle arrays whose densities needs to be
            reinitialized.
        corr : str
            Name of the density reinitialization operation.
            corr='shepard' for using zeroth order shepard filter
        freq : int
            Frequency of reinitialization.
        kernel: any kernel from pysph.base.kernels

        """
        from pysph.solver.utils import get_array_by_name
        self.freq = freq
        self.corr = corr
        self.names = arr_names
        self.count = 1
        self._sph_eval = None
        self.kernel = kernel
        self.dim = app.solver.dim
        self.particles = app.particles
        self.arrs = [get_array_by_name(self.particles, i) for i in self.names]
        options = ['shepard', 'mls2d_1', 'mls3d_1']
        assert self.corr in options, 'corr should be one of %s' % options

    def _get_sph_eval_shepard(self):
        from pysph.sph.wc.density_correction import ShepardFilter
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        if self._sph_eval is None:
            arrs = self.arrs
            eqns = []
            for arr in arrs:
                name = arr.name
                arr.add_property('rhotmp')
                eqns.append(Group(equations=[
                            ShepardFilter(name, [name])], real=False))
            sph_eval = SPHEvaluator(
                arrays=arrs, equations=eqns, dim=self.dim,
                kernel=self.kernel(dim=self.dim))
            return sph_eval
        else:
            return self._sph_eval

    def _get_sph_eval_mls2d_1(self):
        from pysph.sph.wc.density_correction import MLSFirstOrder2D
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        if self._sph_eval is None:
            arrs = self.arrs
            eqns = []
            for arr in arrs:
                name = arr.name
                arr.add_property('rhotmp')
                eqns.append(Group(equations=[
                            MLSFirstOrder2D(name, [name])], real=False))
            sph_eval = SPHEvaluator(
                arrays=arrs, equations=eqns, dim=self.dim,
                kernel=self.kernel(dim=self.dim))
            return sph_eval
        else:
            return self._sph_eval

    def _get_sph_eval_mls3d_1(self):
        from pysph.sph.wc.density_correction import MLSFirstOrder3D
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        if self._sph_eval is None:
            arrs = self.arrs
            eqns = []
            for arr in arrs:
                name = arr.name
                arr.add_property('rhotmp')
                eqns.append(Group(equations=[
                            MLSFirstOrder3D(name, [name])], real=False))
            sph_eval = SPHEvaluator(
                arrays=arrs, equations=eqns, dim=self.dim,
                kernel=self.kernel(dim=self.dim))
            return sph_eval
        else:
            return self._sph_eval

    def _get_sph_eval(self, corr):
        if corr == 'shepard':
            return self._get_sph_eval_shepard()
        elif corr == 'mls2d_1':
            return self._get_sph_eval_mls2d_1()
        elif corr == 'mls3d_1':
            return self._get_sph_eval_mls3d_1()
        else:
            pass

    def post_step(self, solver):
        if self.freq == 0:
            pass
        elif self.count % self.freq == 0:
            self._sph_eval = self._get_sph_eval(self.corr)
            self._sph_eval.update()
            self._sph_eval.evaluate()
        self.count += 1



import numpy as np
import shutil
import os
from os.path import join
import socket
from tempfile import mkdtemp
from pysph import has_h5py

try:
    # This is for Python-2.6.x
    from unittest2 import TestCase, main, skipUnless
except ImportError:
    from unittest import TestCase, main, skipUnless

from pysph.base.utils import get_particle_array, get_particle_array_wcsph
from pysph.solver.utils import dump, load, dump_v1, get_files, get_free_port


class TestGetFiles(TestCase):
    def setUp(self):
        self.root = mkdtemp()
        self.fname = 'dam_break_2d'
        self.dirname = join(self.root, self.fname + '_output')
        os.mkdir(self.dirname)
        self.files = [
            join(
                self.dirname,
                self.fname+'_'+str(i)+'.npz'
            )
            for i in range(11)
        ]
        for name in self.files:
            with open(name, 'w') as fp:
                fp.write('')

    def test_get_files(self):
        self.assertEqual(get_files(self.dirname), self.files)
        self.assertEqual(get_files(self.dirname, fname=self.fname), self.files)
        self.assertEqual(
            get_files(
                self.dirname,
                fname=self.fname,
                endswith=('npz', 'hdf5')
            ),
            self.files
        )

    def tearDown(self):
        shutil.rmtree(self.root)


class TestOutputNumpy(TestCase):
    def setUp(self):
        self.root = mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.root)

    def _get_filename(self, fname):
        return join(self.root, fname) + '.npz'

    def test_dump_and_load_works_by_default(self):
        x = np.linspace(0, 1.0, 10)
        y = x*2.0
        dt = 1.0
        pa = get_particle_array(name='fluid', x=x, y=y)
        fname = self._get_filename('simple')
        dump(fname, [pa], solver_data={'dt': dt})
        data = load(fname)
        solver_data = data['solver_data']
        arrays = data['arrays']
        pa1 = arrays['fluid']
        self.assertListEqual(list(solver_data.keys()), ['dt'])
        self.assertListEqual(list(sorted(pa.properties.keys())),
                             list(sorted(pa1.properties.keys())))
        self.assertTrue(np.allclose(pa.x, pa1.x, atol=1e-14))
        self.assertTrue(np.allclose(pa.y, pa1.y, atol=1e-14))

    def test_dump_and_load_works_with_compress(self):
        x = np.linspace(0, 1.0, 1000)
        y = x*2.0
        dt = 1.0
        pa = get_particle_array(name='fluid', x=x, y=y)
        fname = self._get_filename('simple')
        dump(fname, [pa], solver_data={'dt': dt})
        fnamez = self._get_filename('simplez')
        dump(fnamez, [pa], solver_data={'dt': dt}, compress=True)
        # Check that the file size is indeed smaller
        self.assertTrue(os.stat(fnamez).st_size < os.stat(fname).st_size)

        data = load(fnamez)
        solver_data = data['solver_data']
        arrays = data['arrays']
        pa1 = arrays['fluid']
        self.assertListEqual(list(solver_data.keys()), ['dt'])
        self.assertListEqual(list(sorted(pa.properties.keys())),
                             list(sorted(pa1.properties.keys())))
        self.assertTrue(np.allclose(pa.x, pa1.x, atol=1e-14))
        self.assertTrue(np.allclose(pa.y, pa1.y, atol=1e-14))

    def test_dump_and_load_with_partial_data_dump(self):
        x = np.linspace(0, 1.0, 10)
        y = x*2.0
        pa = get_particle_array_wcsph(name='fluid', x=x, y=y)
        pa.set_output_arrays(['x', 'y'])
        fname = self._get_filename('simple')
        dump(fname, [pa], solver_data={})
        data = load(fname)
        arrays = data['arrays']
        pa1 = arrays['fluid']
        self.assertListEqual(list(sorted(pa.properties.keys())),
                             list(sorted(pa1.properties.keys())))
        self.assertTrue(np.allclose(pa.x, pa1.x, atol=1e-14))
        self.assertTrue(np.allclose(pa.y, pa1.y, atol=1e-14))

    def test_dump_and_load_with_constants(self):
        x = np.linspace(0, 1.0, 10)
        y = x*2.0
        pa = get_particle_array_wcsph(name='fluid', x=x, y=y,
                                      constants={'c1': 1.0, 'c2': [2.0, 3.0]})
        pa.add_property('A', data=2.0, stride=2)
        pa.set_output_arrays(['x', 'y', 'A'])
        fname = self._get_filename('simple')
        dump(fname, [pa], solver_data={})
        data = load(fname)
        arrays = data['arrays']
        pa1 = arrays['fluid']
        self.assertListEqual(list(sorted(pa.properties.keys())),
                             list(sorted(pa1.properties.keys())))
        self.assertListEqual(list(sorted(pa.constants.keys())),
                             list(sorted(pa1.constants.keys())))
        self.assertTrue(np.allclose(pa.x, pa1.x, atol=1e-14))
        self.assertTrue(np.allclose(pa.y, pa1.y, atol=1e-14))
        self.assertTrue(np.allclose(pa.A, pa1.A, atol=1e-14))
        self.assertTrue(np.allclose(pa.c1, pa1.c1, atol=1e-14))
        self.assertTrue(np.allclose(pa.c2, pa1.c2, atol=1e-14))

    def test_that_output_array_information_is_saved(self):
        # Given
        x = np.linspace(0, 1.0, 10)
        y = x*2.0
        pa = get_particle_array(name='fluid', x=x, y=y, u=3*x)

        # When
        output_arrays = ['x', 'y', 'u']
        pa.set_output_arrays(output_arrays)
        fname = self._get_filename('simple')
        dump(fname, [pa], solver_data={})
        data = load(fname)
        pa1 = data['arrays']['fluid']

        # Then.
        self.assertEqual(set(pa.output_property_arrays), set(output_arrays))
        self.assertEqual(set(pa1.output_property_arrays), set(output_arrays))


class TestOutputHdf5(TestOutputNumpy):
    @skipUnless(has_h5py(), "h5py module is not present")
    def setUp(self):
        super(TestOutputHdf5, self).setUp()

    def _get_filename(self, fname):
        return join(self.root, fname) + '.hdf5'


class TestOutputNumpyV1(TestCase):
    def setUp(self):
        self.root = mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.root)

    def _get_filename(self, fname):
        return join(self.root, fname) + '.npz'

    def test_load_works_with_dump_version1(self):
        x = np.linspace(0, 1.0, 10)
        y = x*2.0
        pa = get_particle_array(name='fluid', x=x, y=y)
        fname = self._get_filename('simple')
        dump_v1(fname, [pa], solver_data={})
        data = load(fname)
        arrays = data['arrays']
        pa1 = arrays['fluid']
        self.assertListEqual(list(sorted(pa.properties.keys())),
                             list(sorted(pa1.properties.keys())))
        self.assertTrue(np.allclose(pa.x, pa1.x, atol=1e-14))
        self.assertTrue(np.allclose(pa.y, pa1.y, atol=1e-14))


class TestGetFreePort(TestCase):
    def test_finds_port_not_taken(self):
        # Given
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('', 9800))
        self.addCleanup(sock.close)

        # When
        port = get_free_port(9800)

        # Then
        self.assertNotEqual(port, 9800)
        self.assertTrue(port > 9800)

        # When
        port1 = get_free_port(port)

        # Then
        # getting a free port should not block that port.
        self.assertEqual(port, port1)

    def test_free_port_skips_given(self):
        # Given
        skip = (9800, 9801)

        # When
        port = get_free_port(9800, skip=skip)

        # Then
        self.assertNotIn(port, skip)
        self.assertTrue(port > 9801)


if __name__ == '__main__':
    main()



"""Simulate the classical Sod Shocktube problem in 1D (5 seconds).
"""
from pysph.examples.gas_dynamics.shocktube_setup import ShockTubeSetup
from pysph.sph.scheme import ADKEScheme, GasDScheme, GSPHScheme, SchemeChooser
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.base.nnps import DomainManager

# Numerical constants
dim = 1
gamma = 1.4
gamma1 = gamma - 1.0

# solution parameters
dt = 1e-4
tf = 0.15


class SodShockTube(ShockTubeSetup):

    def initialize(self):
        self.xmin = -0.5
        self.xmax = 0.5
        self.x0 = 0.0
        self.rhol = 1.0
        self.rhor = 0.125
        self.pl = 1.0
        self.pr = 0.1
        self.ul = 0.0
        self.ur = 0.0

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float,
            dest="hdx", default=1.2,
            help="Ratio h/dx."
        )
        group.add_argument(
            "--nl", action="store", type=float, dest="nl", default=640,
            help="Number of particles in left region"
        )

    def consume_user_options(self):
        self.nl = self.options.nl
        self.hdx = self.options.hdx
        ratio = self.rhor/self.rhol
        self.nr = self.nl*ratio
        self.dxl = 0.5/self.nl
        self.dxr = 0.5/self.nr
        self.ml = self.dxl * self.rhol
        self.h0 = self.hdx * self.dxr
        self.hdx = self.hdx
        self.dt = dt
        self.tf = tf

    def create_particles(self):
        # Boundary particles are not needed as we are mirroring the particles
        # using the domain manager. Hence, bx is set to 0.0.
        f, b = self.generate_particles(
            xmin=self.xmin, xmax=self.xmax, dxl=self.dxl, dxr=self.dxr,
            m=self.ml, pl=self.pl, pr=self.pr, h0=self.h0, bx=0.00,
            gamma1=gamma1, ul=self.ul, ur=self.ur
        )
        self.scheme.setup_properties([f, b])
        return [f]

    def create_domain(self):
        return DomainManager(
            xmin=self.xmin, xmax=self.xmax, mirror_in_x=True,
            n_layers=2
        )

    def configure_scheme(self):
        scheme = self.scheme
        if self.options.scheme in ['gsph', 'mpm']:
            scheme.configure(kernel_factor=self.hdx)
        elif self.options.scheme in ['psph', 'tsph']:
            scheme.configure(hfact=self.hdx)
        scheme.configure_solver(tf=self.tf, dt=self.dt)

    def create_scheme(self):
        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=1, beta=1.0, k=0.3, eps=0.5, g1=0.2, g2=0.4)

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=None, alpha1=1.0, alpha2=0.1,
            beta=2.0, update_alpha1=True, update_alpha2=True,
        )
        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=None,
            g1=0.2, g2=0.4, rsolver=2, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=True, blend_alpha=2.0,
            niter=20, tol=1e-6
        )
        crk = CRKSPHScheme(
            fluids=['fluid'], dim=dim, rho0=0, c0=0,
            nu=0, h0=0, p0=0, gamma=gamma, cl=3
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=None
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=None
        )
        s = SchemeChooser(
            default='adke', adke=adke, mpm=mpm, gsph=gsph, crk=crk, psph=psph,
            tsph=tsph)
        return s


if __name__ == '__main__':
    app = SodShockTube()
    app.run()
    app.post_process()



"""Liquid sloshing under horizontal excitation

The case is as described in "ODD M. FALTINSEN, OLAV F. ROGNEBAKKE, IVAN A.
LUKOVSKY, and ALEXANDER N. TIMOKHA. Multidimensional
modal analysis of nonlinear sloshing in a rectangular tank with finite
water depth. Journal of Fluid Mechanics, 407:201234, 2000."

DOI: https://doi.org/10.1017/S0022112099007569


This case is with h=0.6 m and T=1.3 s
"""

# math
from math import cos, pi, exp

import os
import numpy as np

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep
from pysph.sph.integrator import PECIntegrator
from pysph.sph.integrator_step import OneStageRigidBodyStep

from pysph.sph.equation import Group, Equation
from pysph.tools.geometry import get_2d_tank, get_2d_block

from pysph.sph.scheme import WCSPHScheme

Umax = 2*np.sqrt(9.81*0.6)
c0 = 10.0 * Umax

dx = 0.01
hdx = 1.3
h0 = hdx * dx

n_layers = 3
tf = 10.0
rho = 1000.0

alpha = 0.1
beta = 0.0
gamma = 7.0

length = 1.73
h_tank = 1.15
h_liquid = 0.6

amp = 0.032
T = 1.3


class HorizontalExcitation(Equation):
    def __init__(self, dest, sources, amp, T):
        self.amp = amp
        self.T = T
        self.pi = pi
        super(HorizontalExcitation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, t):
        amp = self.amp
        T = self.T
        pi = self.pi

        d_au[d_idx] = -amp * (2*pi/T) * (2*pi/T) * cos(2*pi*t/T)


class SloshingTank(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx', default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx', default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx

    def create_particles(self):
        dx = self.dx
        h0 = self.hdx * self.dx
        m = rho * dx * dx

        xt, yt = get_2d_tank(dx=dx, length=length, height=h_tank,
                             num_layers=n_layers, base_center=[0.0, -dx])

        xf, yf = get_2d_block(dx=dx, length=length - 2*dx,
                              height=h_liquid, center=[0.0, h_liquid*0.5])

        fluid = get_particle_array(name='fluid', x=xf, y=yf, h=h0,
                                   m=m, rho=rho)
        solid = get_particle_array(name='solid', x=xt, y=yt, h=h0,
                                   m=m, rho=rho)

        self.scheme.setup_properties([fluid, solid])
        particles = [fluid, solid]
        return particles

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = PECIntegrator(fluid=WCSPHStep(),
                                   solid=OneStageRigidBodyStep())

        dt = 0.5*self.dx/(1.1*c0)
        self.scheme.configure(h0=self.h0, hdx=self.hdx)

        solver = Solver(kernel=kernel, dim=2, integrator=integrator,
                        tf=tf, dt=dt, adaptive_timestep=True,
                        fixed_h=False)

        return solver

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], ['solid'], dim=2, rho0=rho, c0=c0,
            h0=h0, hdx=hdx, gy=-9.81, alpha=alpha,
            beta=beta, gamma=gamma, hg_correction=True,
            tensile_correction=False
        )
        return s

    def create_equations(self):
        eqns = self.scheme.get_equations()
        equation_1 = Group(
            equations=[
                HorizontalExcitation(
                    dest='solid', sources=None, amp=amp, T=T),
            ], real=False
        )

        eqns.insert(0, equation_1)

        return eqns

    def post_process(self, info_fname):
        """
        Comparing wave height 0.05 m to the left of the tank
        """
        self.read_info(info_fname)

        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        import matplotlib.pyplot as plt
        from cyarray.api import UIntArray
        from pysph.base import nnps

        files = self.output_files
        wave_height, t = [], []
        for sd, arrays1, arrays2 in iter_output(files, 'fluid', 'solid'):
            t.append(sd['t'])
            x_left = arrays2.x.min()
            probe_x = x_left + 0.05 + self.dx * (n_layers - 1)
            probe_y = np.linspace(0.0, arrays1.y.max(), 50)
            probe = get_particle_array(x=probe_x*np.ones_like(probe_y), y=probe_y)
            pa_list = [arrays1, probe]
            nps = nnps.LinkedListNNPS(dim=2, particles=pa_list, radius_scale=1)
            src_index, dst_index = 0, 1
            nps.set_context(src_index=0, dst_index=1)
            nbrs = UIntArray()
            wh = 0.0
            for i in range(len(probe_y)):
                nps.get_nearest_particles(src_index, dst_index, i, nbrs)
                neighbours = nbrs.get_npy_array()
                for j in range(len(neighbours)):
                    if arrays1.y[neighbours[j]] > wh:
                        wh = arrays1.y[neighbours[j]]

            wave_height.append(wh)

        wave_height = np.array(wave_height)
        wave_height = wave_height - h_liquid

        import pysph.examples.st_exp_data as st

        exp_t, exp_H = st.get_faltinsen_data()

        plt.plot(t, wave_height, label="Computed")
        plt.plot(exp_t, exp_H, label="Experiment: Faltinsen et al, 2000")
        plt.xlabel("T (s)")
        plt.ylabel("Wave Height (m)")
        plt.legend()
        plt.savefig(os.path.join(self.output_dir, 'H_vs_t.png'))
        plt.show()


if __name__ == '__main__':
    app = SloshingTank()
    app.run()
    app.post_process(app.info_filename)



from math import exp
from compyle.api import declare
from pysph.sph.equation import Equation
from pysph.sph.gas_dynamics.riemann_solver import (HELPERS, riemann_solve,
                                                   printf)
from pysph.base.particle_array import get_ghost_tag

# Constants
GHOST_TAG = get_ghost_tag()

# Riemann solver types
NonDiffusive = 0
VanLeer = 1
Exact = 2
HLLC = 3
Ducowicz = 4
HLLE = 5
Roe = 6
LLXF = 7
HLLCBall = 8
HLLBall = 9
HLLSY = 10

# GSPHInterpolationType
Delta = 0
Linear = 1
Cubic = 2


def sgn(x=0.0):
    return (x > 0) - (x < 0)


def monotonicity_min(_x1=0.0, _x2=0.0, _x3=0.0):
    x1, x2, x3, _min = declare('double', 4)
    x1 = 2.0 * abs(_x1)
    x2 = abs(_x2)
    x3 = 2.0 * abs(_x3)

    sx1 = sgn(_x1)
    sx2 = sgn(_x2)
    sx3 = sgn(_x3)

    if ((sx1 != sx2) or (sx2 != sx3)):
        return 0.0
    else:
        if x2 < x1:
            if x3 < x2:
                _min = x3
            else:
                _min = x2
        else:
            if x3 < x1:
                _min = x3
            else:
                _min = x1

    return sx1 * _min


class GSPHGradients(Equation):
    def initialize(self, d_idx, d_px, d_py, d_pz, d_ux, d_uy, d_uz,
                   d_vx, d_vy, d_vz, d_wx, d_wy, d_wz):
        d_px[d_idx] = 0.0
        d_py[d_idx] = 0.0
        d_pz[d_idx] = 0.0
        d_ux[d_idx] = 0.0
        d_uy[d_idx] = 0.0
        d_uz[d_idx] = 0.0
        d_vx[d_idx] = 0.0
        d_vy[d_idx] = 0.0
        d_vz[d_idx] = 0.0
        d_wx[d_idx] = 0.0
        d_wy[d_idx] = 0.0
        d_wz[d_idx] = 0.0

    def loop(self, d_idx, d_px, d_py, d_pz, d_ux, d_uy, d_uz,
             d_vx, d_vy, d_vz, d_wx, d_wy, d_wz, d_p, d_u, d_v, d_w,
             s_idx, s_p, s_u, s_v, s_w, s_rho, s_m,
             DWI):
        rj1 = 1.0/s_rho[s_idx]
        pji = s_p[s_idx] - d_p[d_idx]
        uji = s_u[s_idx] - d_u[d_idx]
        vji = s_v[s_idx] - d_v[d_idx]
        wji = s_w[s_idx] - d_w[d_idx]
        tmp = rj1*s_m[s_idx]*pji
        d_px[d_idx] += tmp*DWI[0]
        d_py[d_idx] += tmp*DWI[1]
        d_pz[d_idx] += tmp*DWI[2]
        tmp = rj1*s_m[s_idx]*uji
        d_ux[d_idx] += tmp*DWI[0]
        d_uy[d_idx] += tmp*DWI[1]
        d_uz[d_idx] += tmp*DWI[2]
        tmp = rj1*s_m[s_idx]*vji
        d_vx[d_idx] += tmp*DWI[0]
        d_vy[d_idx] += tmp*DWI[1]
        d_vz[d_idx] += tmp*DWI[2]
        tmp = rj1*s_m[s_idx]*wji
        d_wx[d_idx] += tmp*DWI[0]
        d_wy[d_idx] += tmp*DWI[1]
        d_wz[d_idx] += tmp*DWI[2]


class GSPHUpdateGhostProps(Equation):
    """Copy the GSPH gradients and other props required for GSPH
    from real particle to ghost particles

    """
    def __init__(self, dest, sources=None):
        super(GSPHUpdateGhostProps, self).__init__(dest, sources)
        assert GHOST_TAG == 2

    def initialize(self, d_idx, d_tag, d_orig_idx, d_px, d_py, d_pz,
                   d_ux, d_uy, d_uz, d_vx, d_vy, d_vz, d_wx, d_wy, d_wz,
                   d_grhox, d_grhoy, d_grhoz, d_dwdh, d_rho, d_div,
                   d_p, d_cs):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            # copy pressure grads
            d_px[d_idx] = d_px[idx]
            d_py[d_idx] = d_py[idx]
            d_pz[d_idx] = d_pz[idx]
            # copy u grads
            d_ux[d_idx] = d_ux[idx]
            d_uy[d_idx] = d_uy[idx]
            d_uz[d_idx] = d_uz[idx]
            # copy u grads
            d_vx[d_idx] = d_vx[idx]
            d_vy[d_idx] = d_vy[idx]
            d_vz[d_idx] = d_vz[idx]
            # copy u grads
            d_wx[d_idx] = d_wx[idx]
            d_wy[d_idx] = d_wy[idx]
            d_wz[d_idx] = d_wz[idx]
            # copy density grads
            d_grhox[d_idx] = d_grhox[idx]
            d_grhoy[d_idx] = d_grhoy[idx]
            d_grhoz[d_idx] = d_grhoz[idx]
            # other misc props
            d_dwdh[d_idx] = d_dwdh[idx]
            d_rho[d_idx] = d_rho[idx]
            d_div[d_idx] = d_div[idx]
            d_p[d_idx] = d_p[idx]
            d_cs[d_idx] = d_cs[idx]


class GSPHAcceleration(Equation):
    """Class to implement the GSPH acclerations.

    We implement Inutsuka's original GSPH algorithm I02 defined in
    'Reformulation of Smoothed Particle Hydrodynamics with Riemann
    Solver', (2002), JCP, 179, 238--267 and Iwasaki and Inutsuka's
    vesion (specifically the monotonicity constraint) described in
    'Smoothed particle magnetohydrodynamics with a Riemann solver and
    the method of characteristics' 2011, MNRAS, referred to as IwIn

    Additional details about the algorithm are also described by
    Murante et al.

    """
    def __init__(self, dest, sources, g1=0.0, g2=0.0,
                 monotonicity=0, rsolver=Exact,
                 interpolation=Linear, interface_zero=True, hybrid=False,
                 blend_alpha=5.0, tf=1.0,
                 gamma=1.4, niter=20, tol=1e-6):
        """
        Parameters
        ----------
        g1, g2 : double
            ADKE style thermal conduction parameters
        rsolver: int
            Riemann solver to use.  See pysph.sph.gas_dynamics.gsph for
            valid options.
        interpolation: int
            Kind of interpolation for the specific volume integrals.
        monotonicity : int
            Type of monotonicity algorithm to use:
            0 : First order GSPH
            1 : I02 algorithm
            2 : IwIn algorithm
        interface_zero : bool
            Set Interface position s^*_{ij} = 0 for the Riemann problem.
        hybrid, blend_alpha : bool, double
            Hybrid scheme and blending alpha value
        tf: float
            Final time of simulation for using in blending.
        gamma: float
            Gamma for Equation of state.
        niter: int
            Max number of iterations for iterative Riemann solvers.
        tol: double
            Tolerance for iterative Riemann solvers.
        """
        self.gamma = gamma
        self.niter = niter
        self.tol = tol
        self.g1 = g1
        self.g2 = g2
        self.monotonicity = monotonicity
        self.interpolation = interpolation
        self.rsolver = rsolver
        # Interface position for data reconstruction
        self.sstar = 0.0
        if (g1 == 0 and g2 == 0):
            self.thermal_conduction = 0
        else:
            self.thermal_conduction = 1
        self.interface_zero = interface_zero
        self.hybrid = hybrid
        self.blend_alpha = blend_alpha
        self.tf = tf

        super(GSPHAcceleration, self).__init__(dest, sources)

    def _get_helpers_(self):
        return HELPERS + [monotonicity_min, sgn]

    def initialize(self, d_idx, d_au, d_av, d_aw, d_ae):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0
        d_ae[d_idx] = 0.0

    def loop(self, d_idx, d_m, d_h, d_rho, d_cs, d_div, d_p, d_e, d_grhox,
             d_grhoy, d_grhoz, d_u, d_v, d_w, d_px, d_py, d_pz, d_ux, d_uy,
             d_uz, d_vx, d_vy, d_vz, d_wx, d_wy, d_wz, d_au, d_av, d_aw, d_ae,
             s_idx, s_rho, s_m, s_h, s_cs, s_div, s_p, s_e, s_grhox,
             s_grhoy, s_grhoz, s_u, s_v, s_w, s_px, s_py, s_pz,
             s_ux, s_uy, s_uz, s_vx, s_vy, s_vz, s_wx, s_wy, s_wz,
             XIJ, DWIJ, DWI, DWJ, RIJ, RHOIJ, EPS, dt, t):
        blending_factor = exp(-self.blend_alpha*t/self.tf)
        g1 = self.g1
        g2 = self.g2
        hi = d_h[d_idx]
        hj = s_h[s_idx]
        eij = declare('matrix(3)')
        if RIJ < 1e-14:
            eij[0] = 0.0
            eij[1] = 0.0
            eij[2] = 0.0
            sij = 1.0/(RIJ + EPS)
        else:
            eij[0] = XIJ[0]/RIJ
            eij[1] = XIJ[1]/RIJ
            eij[2] = XIJ[2]/RIJ
            sij = 1.0/RIJ

        vl = s_u[s_idx]*eij[0] + s_v[s_idx]*eij[1] + s_w[s_idx]*eij[2]
        vr = d_u[d_idx]*eij[0] + d_v[d_idx]*eij[1] + d_w[d_idx]*eij[2]

        Hi = g1*hi*d_cs[d_idx] + g2*hi*hi*(abs(d_div[d_idx]) - d_div[d_idx])

        grhoi_dot_eij = (d_grhox[d_idx]*eij[0] + d_grhoy[d_idx]*eij[1]
                         + d_grhoz[d_idx]*eij[2])
        grhoj_dot_eij = (s_grhox[s_idx]*eij[0] + s_grhoy[s_idx]*eij[1]
                         + s_grhoz[s_idx]*eij[2])

        sp_vol = declare('matrix(3)')
        self.interpolate(hi, hj, d_rho[d_idx], s_rho[s_idx], RIJ,
                         grhoi_dot_eij, grhoj_dot_eij, sp_vol)
        vij_i = sp_vol[0]
        vij_j = sp_vol[1]
        sstar = sp_vol[2]
        # Gradients in the local coordinate system
        rsi = (d_grhox[d_idx]*eij[0] + d_grhoy[d_idx]*eij[1] +
               d_grhoz[d_idx]*eij[2])
        psi = (d_px[d_idx]*eij[0] + d_py[d_idx]*eij[1] + d_pz[d_idx]*eij[2])
        vsi = (eij[0]*eij[0]*d_ux[d_idx] +
               eij[0]*eij[1]*(d_uy[d_idx] + d_vx[d_idx]) +
               eij[0]*eij[2]*(d_uz[d_idx] + d_wx[d_idx]) +
               eij[1]*eij[1]*d_vy[d_idx] +
               eij[1]*eij[2]*(d_vz[d_idx] + d_wy[d_idx]) +
               eij[2]*eij[2]*d_wz[d_idx])

        rsj = (s_grhox[s_idx]*eij[0] + s_grhoy[s_idx]*eij[1] +
               s_grhoz[s_idx]*eij[2])
        psj = (s_px[s_idx]*eij[0] + s_py[s_idx]*eij[1] + s_pz[s_idx]*eij[2])
        vsj = (eij[0]*eij[0]*s_ux[s_idx] +
               eij[0]*eij[1]*(s_uy[s_idx] + s_vx[s_idx]) +
               eij[0]*eij[2]*(s_uz[s_idx] + s_wx[s_idx]) +
               eij[1]*eij[1]*s_vy[s_idx] +
               eij[1]*eij[2]*(s_vz[s_idx] + s_wy[s_idx]) +
               eij[2]*eij[2]*s_wz[s_idx])

        csi = d_cs[d_idx]
        csj = s_cs[s_idx]
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        pi = d_p[d_idx]
        pj = s_p[s_idx]

        if self.monotonicity == 0:  # First order scheme
            rsi = 0.0
            rsj = 0.0
            psi = 0.0
            psj = 0.0
            vsi = 0.0
            vsj = 0.0

        if self.monotonicity == 1:  # I02 algorithm
            if (vsi * vsj) < 0:
                vsi = 0.
                vsj = 0.

            # default to first order near a shock
            if (min(csi, csj) < 3.0*(vl - vr)):
                rsi = 0.
                rsj = 0.
                psi = 0.
                psj = 0.
                vsi = 0.
                vsj = 0.

        if self.monotonicity == 2 and RIJ > 1e-14:  # IwIn algorithm
            qijr = rhoi - rhoj
            qijp = pi - pj
            qiju = vr - vl

            delr = rsi * RIJ
            delrp = 2 * delr - qijr
            delp = psi * RIJ
            delpp = 2 * delp - qijp
            delv = vsi * RIJ
            delvp = 2 * delv - qiju

            # corrected values for i
            rsi = monotonicity_min(qijr, delr, delrp)/RIJ
            psi = monotonicity_min(qijp, delp, delpp)/RIJ
            vsi = monotonicity_min(qiju, delv, delvp)/RIJ

            delr = rsj * RIJ
            delrp = 2 * delr - qijr
            delp = psj * RIJ
            delpp = 2 * delp - qijp
            delv = vsj * RIJ
            delvp = 2 * delv - qiju

            # corrected values for j
            rsj = monotonicity_min(qijr, delr, delrp)/RIJ
            psj = monotonicity_min(qijp, delp, delpp)/RIJ
            vsj = monotonicity_min(qiju, delv, delvp)/RIJ
        elif self.monotonicity == 2 and RIJ < 1e-14:  # IwIn algorithm
            rsi = 0.0
            rsj = 0.0
            psi = 0.0
            psj = 0.0
            vsi = 0.0
            vsj = 0.0

        # Input to the riemann solver
        sstar *= 2.0

        # left and right density
        rhol = rhoj + 0.5 * rsj * RIJ * (1.0 - csj*dt*sij + sstar)
        rhor = rhoi - 0.5 * rsi * RIJ * (1.0 - csi*dt*sij + sstar)

        # corrected density
        if rhol < 0:
            rhol = rhoj
        if rhor < 0:
            rhor = rhoi

        # left and right pressure
        pl = pj + 0.5 * psj * RIJ * (1.0 - csj*dt*sij + sstar)
        pr = pi - 0.5 * psi * RIJ * (1.0 - csi*dt*sij + sstar)

        # corrected pressure
        if pl < 0:
            pl = pj
        if pr < 0:
            pr = pi

        # left and right velocity
        ul = vl + 0.5 * vsj * RIJ * (1.0 - csj*dt*sij + sstar)
        ur = vr - 0.5 * vsi * RIJ * (1.0 - csi*dt*sij + sstar)

        # Intermediate state from the Riemann solver
        result = declare('matrix(2)')
        riemann_solve(
            self.rsolver, rhol, rhor, pl, pr, ul, ur,
            self.gamma, self.niter, self.tol, result
        )
        pstar = result[0]
        ustar = result[1]

        # blend of two intermediate states
        if self.hybrid:
            riemann_solve(
                10, rhoj, rhoi, pl, pr, vl, vr, self.gamma,
                self.niter, self.tol, result
            )
            pstar2 = result[0]
            ustar2 = result[1]
            ustar = ustar + blending_factor * (ustar2 - ustar)
            pstar = pstar + blending_factor * (pstar2 - pstar)

        # three dimensional velocity (70)
        vstar = declare('matrix(3)')
        vstar[0] = ustar*eij[0]
        vstar[1] = ustar*eij[1]
        vstar[2] = ustar*eij[2]

        # velocity accelerations
        mj = s_m[s_idx]
        d_au[d_idx] += -mj * pstar * (vij_i * DWI[0] + vij_j * DWJ[0])
        d_av[d_idx] += -mj * pstar * (vij_i * DWI[1] + vij_j * DWJ[1])
        d_aw[d_idx] += -mj * pstar * (vij_i * DWI[2] + vij_j * DWJ[2])

        # contribution to the thermal energy term
        # (85). The contribution due to \dot{x}^*_i will
        # be added in the integrator.
        vstardotdwi = vstar[0]*DWI[0] + vstar[1]*DWI[1] + vstar[2]*DWI[2]
        vstardotdwj = vstar[0]*DWJ[0] + vstar[1]*DWJ[1] + vstar[2]*DWJ[2]
        d_ae[d_idx] += -mj * pstar * (vij_i * vstardotdwi +
                                      vij_j * vstardotdwj)

        # artificial thermal conduction terms
        if self.thermal_conduction:
            divj = s_div[s_idx]
            Hj = g1 * hj * csj + g2 * hj*hj * (abs(divj) - divj)
            Hij = (Hi + Hj) * (d_e[d_idx] - s_e[s_idx])

            Hij /= (RHOIJ * (RIJ*RIJ + EPS))

            d_ae[d_idx] += mj*Hij*(XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] +
                                   XIJ[2]*DWIJ[2])

    def interpolate(self, hi=0.0, hj=0.0, rhoi=0.0, rhoj=0.0,
                    sij=0.0, gri_eij=0.0, grj_eij=0.0,
                    result=[0.0, 0.0, 0.0]):
        """Interpolation for the specific volume integrals in GSPH.

        Parameters:
        -----------

        hi, hj : double
            Particle smoothing length (scale?) at i (right) and j (left)

        rhoi, rhoj : double
            Particle densities at i (right) and j (left)

        sij : double
            Particle separation in the local coordinate system (si - sj)

        gri_eij, grj_eij : double
            Gradient of density at the particles in the global coordinate
            system dot product with eij

        Notes:
        ------

        The interpolation scheme determines the form of the 'specific
        volume' contributions Vij^2 in the GSPH equations.

        The simplest Delta or point interpolation uses Vi^2 =
        1./rho_i^2.

        The more involved linear or cubic spline interpolations are
        defined in the GSPH references.

        Most recent papers on GSPH typically assume the interface to
        be located midway between the particles. In the local
        coordinate system, this corresponds to sij = 0. From I02, it
        seems the definition is only really valid for the constant
        smoothing length case. Set interface_zero to False if you want
        to include this term

        """
        Vi = 1./rhoi
        Vj = 1./rhoj

        Vip = -1./(rhoi*rhoi) * gri_eij
        Vjp = -1./(rhoj*rhoj) * grj_eij

        aij, bij, cij, dij, hij, vij, vij_i, vij_j = declare('double', 8)
        hij = 0.5 * (hi + hj)
        sstar = self.sstar

        # simplest delta or point interpolation
        if self.interpolation == 0:
            vij_i2 = 1./(rhoi * rhoi)
            vij_j2 = 1./(rhoj * rhoj)

        # linear interpolation
        elif self.interpolation == 1:
            # avoid singularities
            if sij < 1e-8:
                cij = 0.0
            else:
                cij = (Vi - Vj)/sij

            dij = 0.5 * (Vi + Vj)

            vij_i2 = 0.25 * hi * hi * cij * cij + dij * dij
            vij_j2 = 0.25 * hj * hj * cij * cij + dij * dij

            # approximate value for the interface location when using
            # variable smoothing lengths
            if not self.interface_zero:
                vij = 0.5 * (vij_i2 + vij_j2)
                sstar = 0.5 * hij*hij * cij*dij/vij

        # cubic spline interpolation
        elif self.interpolation == 2:
            if sij < 1e-8:
                aij = bij = cij = 0.0
                dij = 0.5 * (Vi + Vj)
            else:
                aij = -2.0 * (Vi - Vj)/(sij*sij*sij) + (Vip + Vjp)/(sij*sij)
                bij = 0.5 * (Vip - Vjp)/sij
                cij = 1.5 * (Vi - Vj)/sij - 0.25 * (Vip + Vjp)
                dij = 0.5 * (Vi + Vj) - 0.125*(Vip - Vjp)*sij

            hi2 = hi*hi
            hj2 = hj*hj
            hi4 = hi2*hi2
            hj4 = hj2*hj2
            hi6 = hi4*hi2
            hj6 = hj4*hj2

            vij_i2 = ((15.0)/(64.0)*hi6 * aij*aij +
                      (3.0)/(16.0) * hi4 * (2*aij*cij + bij*bij) +
                      0.25*hi2*(2*bij*dij + cij*cij) + dij * dij)

            vij_j2 = ((15.0)/(64.0)*hj6 * aij*aij +
                      (3.0)/(16.0) * hj4 * (2*aij*cij + bij*bij) +
                      0.25*hj2 * (2*bij*dij + cij*cij) + dij * dij)
            hij2 = hij*hij
            hij4 = hij2*hij2
            if not self.interface_zero:
                vij = 0.5*(vij_i2 + vij_j2)
                sstar = ((15.0/32.0)*hij4*hij2*aij*bij +
                         (3.0/8.0)*hij4*(aij*dij + bij*cij) +
                         0.5*hij2*cij*dij)/vij
        else:
            printf(b"%s", b"Unknown interpolation type")

        result[0] = vij_i2
        result[1] = vij_j2
        result[2] = sstar



"""
Shift particle positions
########################

Equations to maintain uniform particle distribution.

There are two ways proposed:
1. 'SimpleShift' is the earlier one, see [XuStaLau2009].
2. 'FickianShift' is based on Fick's diffusion law, see [LiXuStaRo2012].

TODO: Implement for free surface.

References
----------

    .. [XuStaLau2009] Rui Xu, Peter Stansby, Dominique Laurence "Accuracy
        and stability in incompressible SPH (ISPH) based on the projection
        method and a new approach", Journal of Computational Physics 228
        (2009), pp. 6703--6725.

    .. [LiXuStaRo2012] S.J Lind, R. Xu, P.K. Stansby, B.D. Rogers
        "Incompressible smoothed particle hydrodynamics for free-surface flows:
        A generalised diffusion-based algorithm for stability and validations
        for impulsive flows and propagating waves", Journal of Computational
        Physics 231 (2009), pp. 1499--1523.

    .. [SkLiStaRo2013] Alex Skillen, S. Lind, P.K. Stansby, B.D. Rogers
        "Incompressible smoothed particle hydrodynamics (SPH) with reduced
        temporal noise and generalised Fickian smoothing applied to
        body-water slam and efficient wave-body interaction", Computer Methods
        in Applied Mechanics and Engineering 265 (2013), pp. 163--173.
"""

from math import sqrt
from compyle.api import declare
from pysph.sph.equation import Equation
from pysph.base.reduce_array import parallel_reduce_array, serial_reduce_array
from pysph.solver.tools import Tool


class SimpleShift(Equation):
    r"""**Simple shift**
    See the paper [XuStaLau2009], equation(35)
    """
    def __init__(self, dest, sources, const=0.04):
        self.beta = const
        super(SimpleShift, self).__init__(dest, sources)

    def py_initialize(self, dst, t, dt):
        from numpy import sqrt
        vmag = sqrt(dst.u**2 + dst.v**2 + dst.w**2)
        dst.vmax[0] = serial_reduce_array(vmag, 'max')
        dst.vmax[:] = parallel_reduce_array(dst.vmax, 'max')

    def loop_all(self, d_idx, d_x, d_y, d_z, s_x, s_y, s_z, d_vmax, d_dpos, dt,
                 N_NBRS, NBRS):
        i, s_idx = declare('int', 2)
        ri = 0.0
        dxi = 0.0
        dyi = 0.0
        dzi = 0.0
        eps = 1.0e-08
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij = d_x[d_idx] - s_x[s_idx]
            yij = d_y[d_idx] - s_y[s_idx]
            zij = d_z[d_idx] - s_z[s_idx]
            rij = sqrt(xij*xij + yij*yij + zij*zij)
            r3ij = rij * rij * rij
            dxi += xij / (r3ij + eps)
            dyi += yij / (r3ij + eps)
            dzi += zij / (r3ij + eps)
            ri += rij

        ri = ri/N_NBRS
        fac = self.beta * ri*ri * d_vmax[0] * dt
        d_dpos[d_idx*3] = fac*dxi
        d_dpos[d_idx*3 + 1] = fac*dyi
        d_dpos[d_idx*3 + 2] = fac*dzi

        d_x[d_idx] += d_dpos[d_idx*3]
        d_y[d_idx] += d_dpos[d_idx*3 + 1]
        d_z[d_idx] += d_dpos[d_idx*3 + 2]


class FickianShift(Equation):
    r"""**Fickian-shift**
    See the paper [LiXuStaRo2012], equation(21-24),
    for the constant see [SkLiStaRo2013], equation(13).
    """
    def __init__(self, dest, sources, fickian_const=10, tensile_const=0.2,
                 tensile_pow=4, hdx=1.0, tensile_correction=False):
        self.fickian_const = fickian_const
        self.tensile_const = tensile_const
        self.tensile_pow = tensile_pow
        self.hdx = hdx
        self.tensile_correction = tensile_correction
        super(FickianShift, self).__init__(dest, sources)

    def loop_all(self, d_idx, d_x, d_y, d_z, s_x, s_y, s_z, d_u, d_v, d_w, d_h,
                 s_h, s_m, s_rho, dt, d_dpos,
                 N_NBRS, NBRS, SPH_KERNEL):
        i, s_idx = declare('int', 2)
        xij, dwij, grad_c = declare('matrix(3)', 3)
        grad_c[0] = 0.0
        grad_c[1] = 0.0
        grad_c[2] = 0.0
        ui = d_u[d_idx]
        vi = d_v[d_idx]
        wi = d_w[d_idx]
        vmag = sqrt(ui*ui + vi*vi + wi*wi)

        hi = d_h[d_idx]
        dx = declare('matrix(3)')
        dx[0] = hi/self.hdx
        dx[1] = 0.0
        dx[2] = 0.0
        fij = 0.0
        wdx = SPH_KERNEL.kernel(dx, dx[0], d_h[d_idx])

        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = d_x[d_idx] - s_x[s_idx]
            xij[1] = d_y[d_idx] - s_y[s_idx]
            xij[2] = d_z[d_idx] - s_z[s_idx]
            rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] + xij[2]*xij[2])
            hij = (hi + s_h[s_idx]) * 0.5
            SPH_KERNEL.gradient(xij, rij, hij, dwij)
            Vj = s_m[s_idx] / s_rho[s_idx]

            if self.tensile_correction:
                R = self.tensile_const
                n = self.tensile_pow
                wij = SPH_KERNEL.kernel(xij, rij, hij)
                fij = R * (wij/wdx)**n

            grad_c[0] += Vj * (1 + fij) * dwij[0]
            grad_c[1] += Vj * (1 + fij) * dwij[1]
            grad_c[2] += Vj * (1 + fij) * dwij[2]

        fac = -self.fickian_const * hi * vmag * dt
        d_dpos[d_idx*3] = fac*grad_c[0]
        d_dpos[d_idx*3 + 1] = fac*grad_c[1]
        d_dpos[d_idx*3 + 2] = fac*grad_c[2]

        d_x[d_idx] += d_dpos[d_idx*3]
        d_y[d_idx] += d_dpos[d_idx*3 + 1]
        d_z[d_idx] += d_dpos[d_idx*3 + 2]


class CorrectVelocities(Equation):
    r"""**Correct velocities**
    Correct the velocities after shifting to a new position by using taylor
    series approximation, see equation (34) of [XuStaLau2009].

    .. math::
        \phi_{i}^' = \phi_i + (\nabla \phi)_i \cdot \delta \mathbf{r}_{ii^'}

    where, \phi_{i} is the hydrodynamic variable at old position, \phi_{i}^'
    is at new position, delta \mathbf{r}_{ii^'} is the vector between new and
    old position.
    """
    def initialize(self, d_idx, d_gradv):
        i = declare('int')
        for i in range(9):
            d_gradv[9*d_idx + i] = 0.0

    def loop(self, d_idx, s_idx, s_m, s_rho, d_gradv, DWIJ, VIJ):
        alp, bet = declare('int', 2)

        Vj = s_m[s_idx] / s_rho[s_idx]

        for alp in range(3):
            for bet in range(3):
                d_gradv[d_idx*9 + 3*bet + alp] += -Vj * VIJ[alp] * DWIJ[bet]

    def post_loop(self, d_idx, d_u, d_v, d_w, d_gradv, d_dpos):
        res = declare('matrix(3)')
        i, j = declare('int', 2)
        for i in range(3):
            tmp = 0.0
            for j in range(3):
                tmp += d_gradv[d_idx*9 + 3*i + j] * d_dpos[d_idx*3 + j]
            res[i] = tmp

        d_u[d_idx] += res[0]
        d_v[d_idx] += res[1]
        d_w[d_idx] += res[2]


class ShiftPositions(Tool):
    def __init__(self, app, array_name, freq=1, shift_kind='simple',
                 correct_velocity=False, parameter=None):
        """
        Parameters
        ----------

        app : pysph.solver.application.Application.
            The application instance.
        arr_name : array
            Name of the particle array whose position needs to be
            shifted.
        freq : int
            Frequency to apply particle position shift.
        shift_kind: str
            Kind to shift to apply available are "simple" and "fickian".
        correct_velocity: bool
            Correct velocities after shift in particle position.
        parameter: float
            Correct velocities after shift in particle position.
        """
        from pysph.solver.utils import get_array_by_name
        self.particles = app.particles
        self.dt = app.solver.dt
        self.dim = app.solver.dim
        self.kernel = app.solver.kernel
        self.array = get_array_by_name(self.particles, array_name)
        self.freq = freq
        self.kind = shift_kind
        self.correct_velocity = correct_velocity
        self.parameter = parameter
        self.count = 1
        self._sph_eval = None
        options = ['simple', 'fickian']
        assert self.kind in options, 'shift_kind should be one of %s' % options

    def _get_sph_eval(self, kind):
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        if self._sph_eval is None:
            arr = self.array
            eqns = []
            name = arr.name
            if 'vmax' not in arr.constants.keys():
                arr.add_constant('vmax', [0.0])
            if 'dpos' not in arr.properties.keys():
                arr.add_property('dpos', stride=3)
            if kind == 'simple':
                const = 0.04 if not self.parameter else self.parameter
                eqns.append(Group(
                    equations=[SimpleShift(name, [name], const=const)],
                    update_nnps=True)
                )
            elif kind == 'fickian':
                const = 4 if not self.parameter else self.parameter
                eqns.append(Group(
                    equations=[FickianShift(name, [name],
                                            fickian_const=const)],
                    update_nnps=True)
                )
            if self.correct_velocity:
                if 'gradv' not in arr.properties.keys():
                    arr.add_property('gradv', stride=9)
                eqns.append(Group(equations=[
                    CorrectVelocities(name, [name])]))

            sph_eval = SPHEvaluator(
                arrays=[arr], equations=eqns, dim=self.dim,
                kernel=self.kernel)
            return sph_eval
        else:
            return self._sph_eval

    def post_step(self, solver):
        if self.freq == 0:
            pass
        elif self.count % self.freq == 0:
            self._sph_eval = self._get_sph_eval(self.kind)
            self._sph_eval.update()
            self._sph_eval.evaluate(dt=self.dt)
        self.count += 1



"""1D Rectangular dam break over a wet bed. (5 sec)

The case is described in "Accurate particle splitting for smoothed particle
hydrodynamics in shallow water with shock capturing", R. Vacondio, B. D. Rogers
and P. K. Stansby, Int. J. Numer. Meth. Fluids, Vol 69, pp 1377-1410 (2012).
DOI: 10.1002/fld.2646

"""
import os

# Numpy
from numpy import (ones_like, array, arange, sqrt, concatenate, where, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 1


class RectangularDamBreak(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx1", action="store", type=float, dest="dx1", default=5.0,
            help="Spacing between the particles of first fluid column")
        group.add_argument(
            "--dx2", action="store", type=float, dest="dx2", default=10.0,
            help="Spacing between the particles of second fluid column")
        group.add_argument(
            "--dw01", action="store", type=float, dest="dw01", default=10.0,
            help="Initial depth of the first fluid column (m)")
        group.add_argument(
            "--dw02", action="store", type=float, dest="dw02", default=5.0,
            help="Initial depth of the second fluid column (m)")
        group.add_argument(
            "--l1", action="store", type=float, dest="l1", default=1000.0,
            help="Initial length of the first fluid column (m)")
        group.add_argument(
            "--l2", action="store", type=float, dest="l2", default=1000.0,
            help="Initial length of the second fluid column (m)")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx1 = self.options.dx1
        self.dx2 = self.options.dx2
        self.dw01 = self.options.dw01
        self.dw02 = self.options.dw02
        self.l1 = self.options.l1
        self.l2 = self.options.l2

    def create_particles(self):
        hdx = self.hdx
        dx1 = self.dx1
        dx2 = self.dx2
        l1 = self.l1
        l2 = self.l2
        tot_l = l1 + l2

        d1 = self.dw01
        d2 = self.dw02

        x = concatenate(
                (arange(0, l1, dx1), arange(l1, tot_l+1e-4, dx2)), axis=0
                )

        m = ones_like(x)
        h = ones_like(x)
        h0 = ones_like(x)
        rho = ones_like(x)
        rho0 = ones_like(x)

        # Setting first fluid column properties
        idx_fluid_col_1 = where(x < 1000.0)[0]
        m[idx_fluid_col_1] *= dx1 * rho_w * d1
        h[idx_fluid_col_1] *= hdx * dx1
        h0[idx_fluid_col_1] *= hdx * dx1
        rho[idx_fluid_col_1] *= rho_w * d1
        rho0[idx_fluid_col_1] *= rho_w * d1

        # Setting second fluid column properties
        idx_fluid_col_2 = where(x >= 1000.0)[0]
        m[idx_fluid_col_2] *= dx2 * rho_w * d2
        h[idx_fluid_col_2] *= hdx * dx2
        h0[idx_fluid_col_2] *= hdx * dx2
        rho[idx_fluid_col_2] *= rho_w * d2
        rho0[idx_fluid_col_2] *= rho_w * d2

        fluid = gpa_swe(x=x, m=m, rho=rho, rho0=rho0, h=h, h0=h0,
                        name='fluid')

        # Closed Boundary
        x = concatenate(
                (arange(-2*dx1, l1, dx1), arange(l1, tot_l+2*dx2+1e-4, dx2)),
                axis=0)
        idx_cb_left = where((x < 0))[0]
        idx_cb_right = where((x > 2000))[0]

        m_cb = ones_like(x)
        h_cb = ones_like(x)
        rho_cb = ones_like(x)
        dw_cb = ones_like(x)
        cs_cb = ones_like(x)
        alpha_cb = ones_like(x)

        m_cb[idx_cb_left] *= dx1 * rho_w * d1
        h_cb[idx_cb_left] *= hdx * dx1
        rho_cb[idx_cb_left] *= rho_w * d1
        dw_cb[idx_cb_left] *= d1
        cs_cb[idx_cb_left] *= sqrt(9.8 * dw_cb[idx_cb_left])
        alpha_cb[idx_cb_left] *= dim * rho_cb[idx_cb_left]

        m_cb[idx_cb_right] *= dx2 * rho_w * d2
        h_cb[idx_cb_right] *= hdx * dx2
        rho_cb[idx_cb_right] *= rho_w * d2
        dw_cb[idx_cb_right] *= d2
        cs_cb[idx_cb_right] *= sqrt(9.8 * dw_cb[idx_cb_right])
        alpha_cb[idx_cb_right] *= dim * rho_cb[idx_cb_right]

        boundary = gpa_swe(name='boundary', x=x, m=m_cb,
                           h=h_cb, rho=rho_cb, dw=dw_cb, cs=cs_cb,
                           alpha=alpha_cb)

        idx_to_remove = where((x >= 0) & (x <= 2000))[0]
        boundary.remove_particles(idx_to_remove)

        compute_initial_props([fluid, boundary])
        return [fluid, boundary]

    def create_solver(self):
        kernel = CubicSpline(dim=1)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 60
        solver = Solver(
            kernel=kernel,
            dim=1,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            output_at_times=[10, 20, 30, 40, 50, 60],
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(
                                dest='fluid', sources=['fluid', 'boundary']
                                ),
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid', 'boundary'])
                    ]
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid')
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(
                        dim=dim, dest='fluid', sources=['fluid', 'boundary'],
                        visc_option=2, u_only=True)
                    ],
                ),
            ]
        return equations

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_of_dam()

    def _plot_depth_of_dam(self):
        from matplotlib import pyplot as plt
        t_to_plot = [50.0]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        num = 0
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']
            t = data['solver_data']['t']

            cond1 = fluid.x >= 0
            cond2 = fluid.x <= 2000
            cond = cond1 & cond2
            x = fluid.x[cond]
            dw = fluid.dw[cond]

            x_ind_sort = x.argsort()
            x = x[x_ind_sort]
            x_max = max(x)
            dw = dw[x_ind_sort]
            dw_max = max(self.dw01, self.dw02)
            dw_min = min(self.dw01, self.dw02)

            plt.clf()
            if (
                    (self.dw01 == 10. and self.dw02 == 5.)
                    and (self.l1 == 1000. and self.l2 == 1000.)
                    and abs(t-50.0) < 1e-10
            ):
                x_actual = array(
                        [0., 500., 721.87499, 1468.750, 1468.750, 2000.])
                dw_actual = array([10., 10., 7.27083, 7.27083, 5., 5.])
                plt.plot(x_actual, dw_actual, 'r-', label='Exact Solution',
                         linewidth=2.5)
                fname_res = os.path.join(
                    self.output_dir, 'results.npz')
                savez(
                    fname_res, t=t_to_plot[num], x=x, x_actual=x_actual,
                    dw=dw, dw_actual=dw_actual)

            plt.plot(x, dw, 'bo', label='SWE-SPH', markersize=2)
            plt.ylim(0.8*dw_min, 1.2*dw_max)
            plt.xlim(0, x_max+self.dx2)

            plt.legend(prop={'size': 13})
            plt.xlabel('x (m)', fontsize=13)
            plt.ylabel('dw (m)', fontsize=13)
            plt.tick_params(axis='x', labelsize=13)
            plt.tick_params(axis='y', labelsize=13)
            plt.title('t = %0.1f s' % t_to_plot[num], fontsize=13)
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                SWEOS(dest='fluid')
                ],
            )
    ]
    kernel = CubicSpline(dim=1)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=1,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = RectangularDamBreak()
    app.run()
    app.post_process(app.info_filename)



"""Deformation of a square droplet. (15 minutes)


_______________________________
|                             |
|                             |
|              0              |
|                             |
|        ___________          |
|        |         |          |
|        |    1    |          |
|        |         |          |
|        |_________|          |
|                             |
|                             |
|                             |
|                             |
|_____________________________|



Initially, two two fluids of the same density are distinguished by a
color index assigned to them and allowed to settle under the effects
of surface tension. It is expected that the surface tension at the
interface between the two fluids deforms the initially square droplet
into a cirular droplet to minimize the interface area/length.

The references for this problem are

 - J. Morris "Simulating surface tension with smoothed particle
   hydrodynamics", 2000, IJNMF, 33, pp 333--353 [JM00]

 - S. Adami, X.Y. Hu, N.A. Adams "A new surface tension formulation
   for multi-phase SPH using a reproducing divergence approximation",
   2010, JCP, 229, pp 5011--5021 [AHA10]

 - M. S. Shadloo, M. Yildiz "Numerical modelling of Kelvin-Helmholtz
   instability using smoothed particle hydrodynamics", IJNME, 2011,
   87, pp 988--1006 [SY11]

The surface-tension model used currently is the CSF model based on
interface curvature and normals computed from the color function.

"""

import numpy
import numpy as np
import os

from pysph.sph.surface_tension import get_surface_tension_equations
from pysph.tools.geometry import get_2d_block
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline, QuinticSpline

from pysph.solver.application import Application
from pysph.solver.solver import Solver

from pysph.sph.integrator_step import TransportVelocityStep, \
                        VelocityVerletSymplecticWCSPHStep
from pysph.sph.integrator import PECIntegrator

from pysph.base.nnps import DomainManager
from pysph.solver.utils import iter_output

# problem parameters
dim = 2
domain_width = 1.0
domain_height = 1.0

# numerical constants
sigma = 1.0

# set factor1 to [0.5 ~ 1.0] to simulate a thick or thin
# interface. Larger values result in a thick interface.
factor1 = 0.8
factor2 = 1./factor1

# discretization parameters
dx = dy = 0.0125
dxb2 = dyb2 = 0.5 * dx
volume = dx*dx
hdx = 1.3
h0 = hdx * dx
rho0 = 1
c0 = 20.0
p0 = c0*c0*rho0
nu = 0.2

# correction factor for Morris's Method I. Set with_morris_correction
# to True when using this correction.
epsilon = 0.01/h0

# time steps
tf = 1.0
dt_cfl = 0.25 * h0/(1.1*c0)
dt_viscous = 0.125 * h0**2/nu
dt_force = 1.0

dt = 0.9 * min(dt_cfl, dt_viscous, dt_force)


class SquareDroplet(Application):
    def add_user_options(self, group):
        choices = ['morris', 'tvf', 'adami_stress', 'adami', 'shadloo']
        group.add_argument(
            "--scheme", action="store", dest='scheme', default='morris',
            choices=choices,
            help='Specify scheme to use among %s' % choices
        )

    def create_particles(self):
        x, y = numpy.mgrid[dxb2:domain_width:dx, dyb2:domain_height:dy]
        x = x.ravel()
        y = y.ravel()

        m = numpy.ones_like(x) * volume * rho0
        rho = numpy.ones_like(x) * rho0
        h = numpy.ones_like(x) * h0
        cs = numpy.ones_like(x) * c0

        # additional properties required for the fluid.
        additional_props = [
            # volume inverse or number density
            'V', 'alpha',

            # color and gradients
            'color', 'scolor', 'cx', 'cy', 'cz', 'cx2', 'cy2', 'cz2',

            # discretized interface normals and dirac delta
            'nx', 'ny', 'nz', 'ddelta',

            # interface curvature
            'kappa',

            # transport velocities
            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',

            # imposed accelerations on the solid wall
            'ax', 'ay', 'az', 'wij',

            # velocity of magnitude squared
            'vmag2',

            # variable to indicate reliable normals and normalizing
            # constant
            'N', 'wij_sum', 'pi00', 'pi01', 'pi02', 'pi10', 'pi11', 'pi12',
            'pi20', 'pi21', 'pi22', 'nu'

            ]

        # get the fluid particle array
        fluid = get_particle_array(
            name='fluid', x=x, y=y, h=h, m=m, rho=rho, cs=cs,
            additional_props=additional_props)

        # set the color of the inner square
        for i in range(x.size):
            if ((fluid.x[i] > 0.35) and (fluid.x[i] < 0.65)):
                if ((fluid.y[i] > 0.35) and (fluid.y[i] < 0.65)):
                    fluid.color[i] = 1.0

        # particle volume
        fluid.V[:] = 1./volume
        fluid.nu[:] = nu
        fluid.alpha[:] = sigma
        # set additional output arrays for the fluid
        fluid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                 'ddelta', 'kappa', 'N', 'scolor', 'p'])

        print("2D Square droplet deformation with %d fluid particles" % (
                fluid.get_number_of_particles()))

        return [fluid, ]

    def create_domain(self):
        return DomainManager(
            xmin=0, xmax=domain_width, ymin=0, ymax=domain_height,
            periodic_in_x=True, periodic_in_y=True)

    def create_solver(self):
        kernel = QuinticSpline(dim=2)
        stepper = TransportVelocityStep()
        if self.options.scheme == 'shadloo':
            stepper = VelocityVerletSymplecticWCSPHStep()
        integrator = PECIntegrator(fluid=stepper)
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False)
        return solver

    def create_equations(self):
        return get_surface_tension_equations(['fluid'], [],
                                             self.options.scheme, rho0, p0, c0,
                                             0, factor1, factor2, nu, sigma, 2,
                                             epsilon, 1, real=False)

if __name__ == '__main__':
    app = SquareDroplet()
    app.run()



import threading
import os
import socket
import sys
try:
    from SimpleXMLRPCServer import (SimpleXMLRPCServer,
                                    SimpleXMLRPCRequestHandler)
    from SimpleHTTPServer import SimpleHTTPRequestHandler
except ImportError:
    # Python 3.x
    from xmlrpc.server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
    from http.server import SimpleHTTPRequestHandler

from multiprocessing.managers import BaseManager, BaseProxy


def get_authkey_bytes(authkey):
    if isinstance(authkey, bytes):
        return authkey
    else:
        return authkey.encode('utf-8')


class MultiprocessingInterface(BaseManager):
    """ A multiprocessing interface to the solver controller

    This object exports a controller instance proxy over the multiprocessing
    interface. Control actions can be performed by connecting to the interface
    and calling methods on the controller proxy instance """
    def __init__(self, address=None, authkey=None):
        authkey = get_authkey_bytes(authkey)
        super().__init__(address, authkey)
        self.authkey = authkey
        self._server = None

    def stop(self):
        if self._server:
            conn = self._Client(self._address, authkey=self._authkey)
            try:
                self._server.shutdown(conn)
            finally:
                conn.close()
        elif hasattr(self, 'shutdown'):
            self.shutdown()

    def get_controller(self):
        return self.controller

    def start(self, controller):
        self.controller = controller
        self.register('get_controller', self.get_controller)
        if sys.platform == 'win32':
            self.start()
        else:
            self._server = self.get_server()
            self._server.serve_forever()


class MultiprocessingClient(BaseManager):
    """ A client for the multiprocessing interface

    Override the run() method to do appropriate actions on the proxy
    instance of the controller object or add an interface using the
    add_interface methods similar to the Controller.add_interface method """
    def __init__(self, address=None, authkey=None, serializer='pickle',
                 start=True):
        authkey = get_authkey_bytes(authkey)
        BaseManager.__init__(self, address, authkey, serializer)
        if start:
            self.start()

    def start(self, connect=True):
        self.interfaces = []

        # to work around a python caching bug
        # http://stackoverflow.com/questions/3649458/broken-pipe-when-using-python-multiprocessing-managers-basemanager-syncmanager
        if self.address in BaseProxy._address_to_local:
            del BaseProxy._address_to_local[self.address][0].connection

        self.register('get_controller')
        if connect:
            self.connect()
            self.controller = self.get_controller()
        self.run(self.controller)

    @staticmethod
    def is_available(address):
        try:
            socket.create_connection(address, 1).close()
            return True
        except socket.error:
            return False

    def run(self, controller):
        pass

    def add_interface(self, callable):
        """ This makes it act as substitute for the actual command_manager """
        thr = threading.Thread(target=callable, args=(self.controller,))
        thr.daemon = True
        thr.start()
        return thr


class CrossDomainXMLRPCRequestHandler(SimpleXMLRPCRequestHandler,
                                      SimpleHTTPRequestHandler):
    """ SimpleXMLRPCRequestHandler subclass which attempts to do CORS

    CORS is Cross-Origin-Resource-Sharing (http://www.w3.org/TR/cors/)
    which enables xml-rpc calls from a different domain than the xml-rpc server
    (such requests are otherwise denied)
    """
    def do_OPTIONS(self):
        """ Implement the CORS pre-flighted access for resources """
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-METHODS", "POST,GET,OPTIONS")
        # self.send_header("Access-Control-Max-Age", "60")
        self.send_header("Content-length", "0")
        self.end_headers()

    def do_GET(self):
        """ Handle http requests to serve html/image files only """
        print(self.path, self.translate_path(self.path))
        permitted_extensions = ['.html', '.png', '.svg', '.jpg', '.js']
        if not os.path.splitext(self.path)[1] in permitted_extensions:
            self.send_error(404, 'File Not Found/Allowed')
        else:
            SimpleHTTPRequestHandler.do_GET(self)

    def end_headers(self):
        """ End response header with adding Access-Control-Allow-Origin

        This is done to enable CORS request from all clients """
        self.send_header("Access-Control-Allow-Origin", "*")
        SimpleXMLRPCRequestHandler.end_headers(self)


class XMLRPCInterface(SimpleXMLRPCServer):
    """ An XML-RPC interface to the solver controller

    Currently cannot work with objects which cannot be marshalled
    (which is basically most custom classes, most importantly
    ParticleArray and numpy arrays) """
    def __init__(self, addr, requestHandler=CrossDomainXMLRPCRequestHandler,
                 logRequests=True, allow_none=True,
                 encoding=None, bind_and_activate=True):
        SimpleXMLRPCServer.__init__(self, addr, requestHandler, logRequests,
                                    allow_none, encoding, bind_and_activate)

    def stop(self):
        self.server_close()

    def start(self, controller):
        self.register_instance(controller, allow_dotted_names=False)
        self.register_introspection_functions()
        self.serve_forever()


class CommandlineInterface(object):
    """ command-line interface to the solver controller """
    def start(self, controller):
        while True:
            try:
                try:
                    inp = raw_input('pysph[%d]>>> ' % controller.get('count'))
                except NameError:
                    inp = input('pysph[%d]>>> ' % controller.get('count'))
                cmd = inp.strip().split()
                try:
                    cmd, args = cmd[0], cmd[1:]
                except Exception as e:
                    print('Invalid command')
                    self.help()
                    continue
                args2 = []
                for arg in args:
                    try:
                        arg = eval(arg)
                    except:
                        pass
                    finally:
                        args2.append(arg)

                if cmd == 'p' or cmd == 'pause':
                    controller.pause_on_next()
                elif cmd == 'c' or cmd == 'cont':
                    controller.cont()
                elif cmd == 'g' or cmd == 'get':
                    print(controller.get(args[0]))
                elif cmd == 's' or cmd == 'set':
                    print(controller.set(args[0], args2[1]))
                elif cmd == 'q' or cmd == 'quit':
                    break
                else:
                    print(getattr(controller, cmd)(*args2))
            except Exception as e:
                self.help()
                print(e)

    def help(self):
        print('''Valid commands are:
    p | pause
    c | cont
    g | get <name>
    s | set <name> <value>
    q | quit -- quit commandline interface (solver keeps running)''')
